#include "precomp.h"

/* 0 = Only log errors.
   1 = log debug information.
 */
int aslcxp_device_debug = 0;

static int aslcxp_open(struct inode *inode, struct file *file)
{
   struct aslcxp_device *cxp_device = NULL;
   int error = -EIO;

   cxp_device = container_of(inode->i_cdev, struct aslcxp_device, cdev);

   mutex_lock(&cxp_device->open_mutex);
   if ((file->f_flags & O_EXCL) && (cxp_device->open_count > 0)){
      dev_err(cxp_device->device, "%s: Device is busy\n", __FUNCTION__);
      error = -EBUSY;
   }
   else{
      if (cxp_device->open_count == 0) aslcxp_hw_start(cxp_device);
      cxp_device->open_count++;
      if (aslcxp_device_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: open_count = %d\n", __FUNCTION__,  cxp_device->open_count);
      error = 0;
   }
   mutex_unlock(&cxp_device->open_mutex);

   if (error == 0) file->private_data = cxp_device;

   return error;
}

static int aslcxp_release(struct inode *inode, struct file *file)
{
   struct aslcxp_device *cxp_device = NULL;
   int error = -EIO;

   cxp_device = container_of(inode->i_cdev, struct aslcxp_device, cdev);

   mutex_lock(&cxp_device->open_mutex);
   if (cxp_device->open_count > 0){
      cxp_device->open_count--;
      if (aslcxp_device_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: open_count = %d\n", __FUNCTION__,  cxp_device->open_count);
      if (cxp_device->open_count == 0) aslcxp_hw_reset(cxp_device);
      error = 0;
   }
   else{
      error = -ENOENT;
   }
   mutex_unlock(&cxp_device->open_mutex);

   if (error == -ENOENT) dev_err(cxp_device->device, "%s: Device not open\n", __FUNCTION__);

   if (error == 0){
      file->private_data = NULL;
   }

   return error;
}

static void aslcxp_ioctl_new_op(struct aslcxp_device *cxp_device)
{
   cxp_device->cxp_transaction.error_code = 0;
   cxp_device->cxp_transaction.ioctl = 0;
   cxp_device->cxp_transaction.protocol_error_code = 0;
   memset(&cxp_device->cxp_transaction.cxpheader, 0, sizeof(cxp_device->cxp_transaction.cxpheader));
   if (cxp_device->cxp_transaction.cxpdata) kfree(cxp_device->cxp_transaction.cxpdata);
   cxp_device->cxp_transaction.cxpdata = NULL;
   return;
}

static int aslcxp_ioctl_process_reset(struct aslcxp_device *cxp_device, mda_io_buffer *user_io)
{
   int error = 0;

   mutex_lock(&cxp_device->cxp_serialisation_mutex);

   aslcxp_ioctl_new_op(cxp_device);
   cxp_device->cxp_transaction.ioctl = MDACXP_IOCTL_RESET_DEVICE;
   error = aslcxp_hw_start_op(cxp_device);
   if (error){
      dev_err(cxp_device->device, "%s: Failed with %d\n", __FUNCTION__, error);
      goto bail;
   }

bail:
   mutex_unlock(&cxp_device->cxp_serialisation_mutex);

   user_io->output_buffer.user_size = 0;
   return error;
}

static int aslcxp_ioctl_process_read(struct aslcxp_device *cxp_device, mda_io_buffer *user_io)
{
   int error = 0;
   size_t output_structure_size = 0;

   mutex_lock(&cxp_device->cxp_serialisation_mutex);

   aslcxp_ioctl_new_op(cxp_device);
   cxp_device->cxp_transaction.ioctl = MDACXP_IOCTL_CXP_READ;

   if (user_io->param_buffer.user_size < sizeof(cxp_device->cxp_transaction.cxpheader))
   {
	   dev_err(cxp_device->device, "%s: Parameter buffer too small\n", __FUNCTION__);
	   error = -EINVAL;
	   goto bail;
   }
   error = copy_from_user(&cxp_device->cxp_transaction.cxpheader, user_io->param_buffer.user_address, sizeof(cxp_device->cxp_transaction.cxpheader));
   if (error){
      dev_err(cxp_device->device, "%s: Could not copy parameter block in\n", __FUNCTION__);
      goto bail;
   }
   output_structure_size = sizeof(cxp_device->cxp_transaction.cxpheader) + cxp_device->cxp_transaction.cxpheader.dwByteCount;
   if (aslcxp_device_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: %d byte read requested\n", __FUNCTION__, cxp_device->cxp_transaction.cxpheader.dwByteCount);
   if (user_io->output_buffer.user_size < output_structure_size){
      dev_err(cxp_device->device, "%s: Output buffer too small\n", __FUNCTION__);
      error = -EINVAL;
      goto bail;
   }
   cxp_device->cxp_transaction.cxpdata = kzalloc(output_structure_size, GFP_KERNEL);
   if (!cxp_device->cxp_transaction.cxpdata){
      dev_err(cxp_device->device, "%s: Allocation (%u) failed\n", __FUNCTION__, (uint32_t) output_structure_size);
      error = -ENOMEM;
      goto bail;
   }
   memcpy(cxp_device->cxp_transaction.cxpdata, &cxp_device->cxp_transaction.cxpheader, sizeof(cxp_device->cxp_transaction.cxpheader));

   error = aslcxp_hw_start_op(cxp_device);
   if (error){
      dev_err(cxp_device->device, "%s: Failed with %d\n", __FUNCTION__, error);
      goto bail;
   }

   error = copy_to_user(user_io->output_buffer.user_address, cxp_device->cxp_transaction.cxpdata, output_structure_size);
   if (error){
      dev_err(cxp_device->device, "%s: Failed to copy buffer out\n", __FUNCTION__);
      goto bail;
   }
   user_io->output_buffer.user_size = output_structure_size;

bail:
   if (cxp_device->cxp_transaction.cxpdata) kfree(cxp_device->cxp_transaction.cxpdata);
   cxp_device->cxp_transaction.cxpdata = NULL;

   mutex_unlock(&cxp_device->cxp_serialisation_mutex);

   if (error) user_io->output_buffer.user_size = 0;
   return error;
}

static int aslcxp_ioctl_process_write(struct aslcxp_device *cxp_device, mda_io_buffer *user_io)
{
   int error = 0;
   size_t input_structure_size = 0;

   mutex_lock(&cxp_device->cxp_serialisation_mutex);

   aslcxp_ioctl_new_op(cxp_device);
   cxp_device->cxp_transaction.ioctl = MDACXP_IOCTL_CXP_WRITE;

   if (user_io->param_buffer.user_size < sizeof(cxp_device->cxp_transaction.cxpheader))
   {
     dev_err(cxp_device->device, "%s: Parameter buffer too small\n", __FUNCTION__);
     error = -EINVAL;
     goto bail;
   }
   error = copy_from_user(&cxp_device->cxp_transaction.cxpheader, user_io->param_buffer.user_address, sizeof(cxp_device->cxp_transaction.cxpheader));
   if (error){
      dev_err(cxp_device->device, "%s: Could not copy parameter block in\n", __FUNCTION__);
      goto bail;
   }
   input_structure_size = sizeof(cxp_device->cxp_transaction.cxpheader) + cxp_device->cxp_transaction.cxpheader.dwByteCount;
   if (aslcxp_device_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: %d byte write request (%d)\n", __FUNCTION__, cxp_device->cxp_transaction.cxpheader.dwByteCount, (int)input_structure_size);
   if (user_io->input_buffer.user_size < input_structure_size){
		   dev_err(cxp_device->device, "%s: Output buffer too small\n", __FUNCTION__);
		   error = -EINVAL;
		   goto bail;
   }
   cxp_device->cxp_transaction.cxpdata = kzalloc(input_structure_size, GFP_KERNEL);
   if (!cxp_device->cxp_transaction.cxpdata){
		   dev_err(cxp_device->device, "%s: Allocation (%u) failed\n", __FUNCTION__, (uint32_t) input_structure_size);
		   error = -ENOMEM;
		   goto bail;
   }

   error = copy_from_user(cxp_device->cxp_transaction.cxpdata, user_io->input_buffer.user_address, input_structure_size);
   if (error){
      dev_err(cxp_device->device, "%s: Failed to copy buffer in\n", __FUNCTION__);
      goto bail;
   }

   error = aslcxp_hw_start_op(cxp_device);
   if (error){
      dev_err(cxp_device->device, "%s: Failed with %d\n", __FUNCTION__, error);
      goto bail;
   }

bail:
   if (cxp_device->cxp_transaction.cxpdata) kfree(cxp_device->cxp_transaction.cxpdata);
   cxp_device->cxp_transaction.cxpdata = NULL;

   mutex_unlock(&cxp_device->cxp_serialisation_mutex);

   return error;
}

static int aslcxp_ioctl_process_last_error(struct aslcxp_device *cxp_device, mda_io_buffer *user_io)
{
   int error = 0;
   uint32_t cxp_last_protocol_error = 0;

   if (user_io->output_buffer.user_size < sizeof(cxp_last_protocol_error)){
      dev_err(cxp_device->device, "%s: Output buffer too small", __FUNCTION__);
      error = -EINVAL;
      goto bail;
   }

   mutex_lock(&cxp_device->cxp_serialisation_mutex);
   cxp_last_protocol_error = cxp_device->cxp_last_protocol_error;
   cxp_device->cxp_last_protocol_error = 0;
   mutex_unlock(&cxp_device->cxp_serialisation_mutex);

   error = copy_to_user(user_io->output_buffer.user_address, &cxp_last_protocol_error, sizeof(cxp_last_protocol_error));
   if (aslcxp_device_debug && !error) dev_printk(KERN_INFO, cxp_device->device, "%s: last_protocol_error = %08X\n", __FUNCTION__, cxp_last_protocol_error);

bail:
   if (error) user_io->output_buffer.user_size = 0;
   else user_io->output_buffer.user_size = sizeof(cxp_last_protocol_error);
   return error;
}

static long aslcxp_ioctl_process(struct file *file, unsigned int cmd, mda_io_buffer *user_io)
{
   int error = 0;
   struct aslcxp_device * cxp_device = file->private_data;

   switch(cmd){
   case MDACXP_IOCTL_RESET_DEVICE:
      error = aslcxp_ioctl_process_reset(cxp_device, user_io);
      break;
   case MDACXP_IOCTL_CXP_READ:
      error = aslcxp_ioctl_process_read(cxp_device, user_io);
      break;
   case MDACXP_IOCTL_CXP_WRITE:
      error = aslcxp_ioctl_process_write(cxp_device, user_io);
      break;
   case MDACXP_IOCTL_CXP_GET_LAST_ERROR:
      error = aslcxp_ioctl_process_last_error(cxp_device, user_io);
      break;
   case MDACXP_IOCTL_CXP_SET_ACK_TIMEOUT: /* CLR999 TODO */
      user_io->output_buffer.user_size = 0;
      error = 0;
      break;
   case MDA_IOCTL_GET_VERSION:
      error = mda_ioctl_version(user_io, ASLCXP_VERSION);
      break;
   case MDA_IOCTL_GET_UNIQUE_ID:
      error = mda_ioctl_unique_id(user_io, cxp_device->mdadev);
      break;
   case MDA_IOCTL_SYSTEM_REGLIST:
      error =  mda_ioctl_reglist(cxp_device->mdadev, user_io, 1);
      break;
   default:
      error = -ENOTSUPP;
      break;
   }

   if ((error) && (error != -ENOTSUPP)) dev_err(cxp_device->device, "%s: returning %d\n", __FUNCTION__, (int)error);
   return (long)error;
}

static long aslcxp_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
   long error = -EINVAL;
   mda_io_buffer user64_io_buffer;

   memset(&user64_io_buffer, 0, sizeof(user64_io_buffer));
   if (arg != 0){
      error = copy_from_user(&user64_io_buffer, (void*)arg, sizeof(user64_io_buffer));
      if (error) goto bail;
      error = aslcxp_ioctl_process(file, cmd, &user64_io_buffer);
   }
bail:
   if (!error){
      error = copy_to_user((void*)arg, &user64_io_buffer, sizeof(user64_io_buffer));
   }
   return error;
}

#if defined __LP64__

static long aslcxp_compat(struct file *file, unsigned int cmd, unsigned long arg)
{
   long error = -EINVAL;
   mda_io_buffer user64_io_buffer;
   struct mda_io32_buffer user32_io_buffer;

   if (arg != 0){
      error = copy_from_user(&user32_io_buffer, (void*)arg, sizeof(user32_io_buffer));
      if (error) goto bail;
      user64_io_buffer.param_buffer.user_address = (void*)(u64)user32_io_buffer.param_buffer.user_address;
      user64_io_buffer.param_buffer.user_size = (size_t)user32_io_buffer.param_buffer.user_size;
      user64_io_buffer.input_buffer.user_address = (void*)(u64)user32_io_buffer.input_buffer.user_address;
      user64_io_buffer.input_buffer.user_size = (size_t)user32_io_buffer.input_buffer.user_size;
      user64_io_buffer.output_buffer.user_address = (void*)(u64)user32_io_buffer.output_buffer.user_address;
      user64_io_buffer.output_buffer.user_size = (size_t)user32_io_buffer.output_buffer.user_size;
      error = aslcxp_ioctl_process(file, cmd, &user64_io_buffer);
   }
bail:
   if (!error){
      user32_io_buffer.output_buffer.user_size = (uint32_t) user64_io_buffer.output_buffer.user_size;
      error = copy_to_user((void*)arg, &user32_io_buffer, sizeof(user32_io_buffer));
   }
   return error;
}

#endif

struct file_operations aslcxp_ops = {
   .owner = THIS_MODULE,
   .open  = aslcxp_open,
   .release = aslcxp_release,
   .unlocked_ioctl = aslcxp_ioctl,
#if defined __LP64__
   .compat_ioctl = aslcxp_compat,
#endif
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
MODULE_DEVICE_TABLE(file_operations, aslcxp_ops);
#else
MODULE_GENERIC_TABLE(file_operations, aslcxp_ops);
#endif

