#include "precomp.h"

/* 0 = Only log errors.
 * 1 = High level logging.
 * 2 = Register level logging
 */
int aslcxp_hw_debug = 0;

void aslcxp_complete_transaction(struct cxp_transaction *transaction, long error_code, uint32_t protocol_error_code)
{
   unsigned long irq_state = 0;
   int wake_sleeper = 0;

   spin_lock_irqsave(&transaction->cxp_device->interrupt_spinlock, irq_state);
   wake_sleeper = aslcxp_complete_transaction_locked(transaction, error_code, protocol_error_code);
   spin_unlock_irqrestore(&transaction->cxp_device->interrupt_spinlock, irq_state);

   if (wake_sleeper) wake_up(&transaction->wait_queue_head);

   return;
}

void aslcxp_hw_reset(struct aslcxp_device *cxp_device)
{
   unsigned long irq_state = 0;

   spin_lock_irqsave(&cxp_device->interrupt_spinlock, irq_state);
   aslcxp_hw_reset_locked(cxp_device);
   spin_unlock_irqrestore(&cxp_device->interrupt_spinlock, irq_state);
   return;
}

int aslcxp_link_is_active(struct aslcxp_device *cxp_device)
{
   unsigned long irq_state = 0;
   uint32_t dl_status = 0;

   spin_lock_irqsave(&cxp_device->interrupt_spinlock, irq_state);
   dl_status = aslcxp_read32(cxp_device, FBD_CXP_ADDR_HOST_DL_STATUS);
   spin_unlock_irqrestore(&cxp_device->interrupt_spinlock, irq_state);
   return (dl_status & FBD_CXP_VALUE_HOST_DL_STATUS_LINK_ACTIVE) ? 1 : 0;
}

void aslcxp_hw_start(struct aslcxp_device *cxp_device)
{
   unsigned long irq_state = 0;

   spin_lock_irqsave(&cxp_device->interrupt_spinlock, irq_state);
   aslcxp_hw_start_locked(cxp_device);
   spin_unlock_irqrestore(&cxp_device->interrupt_spinlock, irq_state);
   return;

}

void aslcxp_timeout_callback(struct cxp_transaction *transaction)
{
   unsigned long irq_state = 0;
   uint32_t protocol_error_code = 0;
   int wake_sleeper = 0;

   if (aslcxp_hw_debug && transaction->cxp_device) dev_printk(KERN_INFO, transaction->cxp_device->device, "%s\n", __FUNCTION__);

   spin_lock_irqsave(&transaction->cxp_device->interrupt_spinlock, irq_state);
   switch(transaction->cxp_transaction_state){
      case cxp_started:
         protocol_error_code = FBD_CXP_ASL_VALUE_TO_HOST_ERROR_ACK_TIMEOUT;
         break;
      case cxp_tentative_ack:
         protocol_error_code = FBD_CXP_ASL_VALUE_TO_HOST_ERROR_TENTATIVE_ACK_TIMEOUT;
         break;
      case cxp_wait_ack:
         protocol_error_code = FBD_CXP_ASL_VALUE_TO_HOST_ERROR_WAIT_ACK_TIMEOUT;
         break;
      default:
         break;
   }
   if (protocol_error_code){
      wake_sleeper = aslcxp_complete_transaction_locked(transaction, -ETIME, protocol_error_code);
      if (aslcxp_hw_debug) dev_printk(KERN_INFO, transaction->cxp_device->device, "%s: protocol error 0x%08X\n", __FUNCTION__, protocol_error_code);
   }
   spin_unlock_irqrestore(&transaction->cxp_device->interrupt_spinlock, irq_state);

   if (wake_sleeper){
      wake_up(&transaction->wait_queue_head);
      if (aslcxp_hw_debug) dev_printk(KERN_INFO, transaction->cxp_device->device, "%s: waking sleeper\n", __FUNCTION__);
   }
   return;
}

int aslcxp_hw_start_op(struct aslcxp_device *cxp_device)
{
   int error = 0;
   int locked = 0;
   uint32_t dl_status = 0;
   uint32_t ul_status = 0;
   uint32_t fifo_bytes_free = 0;
   uint32_t fifo_entry = 0;
   uint32_t fifo_bytes_written = 0;
   uint32_t cxp_command = 0;
   uint32_t cxp_byte_count = 0;
   DEFINE_WAIT(wait);
   unsigned long irq_state = 0;
   int timeout_ok = 0;

   dl_status = aslcxp_read32(cxp_device, FBD_CXP_ADDR_HOST_DL_STATUS);
   if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: dl_status = %08X\n", __FUNCTION__, dl_status);

   if (0 == (dl_status & FBD_CXP_VALUE_HOST_DL_STATUS_LINK_ACTIVE))
   {
      dev_err(cxp_device->device, "%s: Link inactive\n", __FUNCTION__);
      error = -ENXIO;
      goto bail;
   }
   if (0 != (dl_status & FBD_CXP_VALUE_HOST_DL_STATUS_ACK_PENDING))
   {
      dev_err(cxp_device->device, "%s: Link busy: dl_status = %08X\n", __FUNCTION__, dl_status);
      error = -EBUSY;
      goto bail;
   }
   if (0 != (dl_status & FBD_CXP_VALUE_HOST_DL_STATUS_FIFO_OVERFLOW))
   {
      dev_err(cxp_device->device, "%s: FIFO overflowed\n", __FUNCTION__);
      error = -EMSGSIZE;
      goto bail;
   }

   spin_lock_irqsave(&cxp_device->interrupt_spinlock, irq_state);
   locked = 1;

   if (cxp_device->cxp_transaction.cxp_transaction_state != cxp_idle)
   {
      dev_err(cxp_device->device, "%s: Transaction outstanding\n", __FUNCTION__);
      error = -EBUSY;
      goto bail;
   }

   switch(cxp_device->cxp_transaction.ioctl)
   {
   case MDACXP_IOCTL_CXP_WRITE:
      if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: write\n", __FUNCTION__);
      if (cxp_device->cxp_transaction.cxpheader.dwByteCount > FBD_CXP_MASK_TO_DEVICE_COMMAND_BYTE_COUNT)
      {
         error = -EINVAL;
         goto bail;
      }
      ul_status = aslcxp_read32(cxp_device, FBD_CXP_ADDR_HOST_UL_STATUS);
      fifo_bytes_free = sizeof(fifo_entry) * (ul_status & FBD_CXP_MASK_HOST_UL_STATUS_FIFO_SPACES);
      if (fifo_bytes_free < cxp_device->cxp_transaction.cxpheader.dwByteCount)
      {
         error = -EINVAL;
         goto bail;
      }
      fifo_entry = 0;
      for(fifo_bytes_written = 0; fifo_bytes_written < cxp_device->cxp_transaction.cxpheader.dwByteCount; fifo_bytes_written++)
      {
         uint32_t dwByteNumber = fifo_bytes_written % sizeof(fifo_entry);
         uint32_t dwShift = 8 * (sizeof(fifo_entry) - (dwByteNumber + 1));
         fifo_entry |= cxp_device->cxp_transaction.cxpdata->baData[fifo_bytes_written] << dwShift;
         if (dwByteNumber == (sizeof(fifo_entry) - 1)) /* The last byte of a word */
         {
            aslcxp_write32(cxp_device, FBD_CXP_ADDR_TO_DEVICE_FIFO, fifo_entry);
            if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: wrote 0x%08X to FIFO\n", __FUNCTION__, fifo_entry);
            fifo_entry = 0;
         }
      }
      if (fifo_bytes_written % sizeof(fifo_entry)) /* Need to write out the remainder */
      {
         aslcxp_write32(cxp_device, FBD_CXP_ADDR_TO_DEVICE_FIFO, fifo_entry);
         if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: wrote 0x%08X to FIFO\n", __FUNCTION__, fifo_entry);
      }
      cxp_command = FBD_CXP_VALUE_TO_DEVICE_COMMAND_WRITE;
      cxp_byte_count = cxp_device->cxp_transaction.cxpheader.dwByteCount;
      aslcxp_write32(cxp_device, FBD_CXP_ADDR_TO_DEVICE_ADDRESS, cxp_device->cxp_transaction.cxpheader.dwDeviceAddress);
      if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: wrote 0x%08X to address\n", __FUNCTION__, cxp_device->cxp_transaction.cxpheader.dwDeviceAddress);
      break;
   case MDACXP_IOCTL_CXP_READ:
      if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: read\n", __FUNCTION__);
      if (cxp_device->cxp_transaction.cxpheader.dwByteCount > FBD_CXP_MASK_TO_DEVICE_COMMAND_BYTE_COUNT)
      {
         error = -EINVAL;
         goto bail;
      }
      cxp_command = FBD_CXP_VALUE_TO_DEVICE_COMMAND_READ;
      cxp_byte_count = cxp_device->cxp_transaction.cxpheader.dwByteCount;
      aslcxp_write32(cxp_device, FBD_CXP_ADDR_TO_DEVICE_ADDRESS, cxp_device->cxp_transaction.cxpheader.dwDeviceAddress);
      if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: wrote 0x%08X to address\n", __FUNCTION__, cxp_device->cxp_transaction.cxpheader.dwDeviceAddress);
      break;
   case MDACXP_IOCTL_RESET_DEVICE:
      if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: reset\n", __FUNCTION__);
      cxp_command = FBD_CXP_VALUE_TO_DEVICE_COMMAND_RESET;
      cxp_byte_count = 0;
      timeout_ok = 1;
      break;
   default:
      error = -ENOTSUPP;
      goto bail;
      break;
   }

   while(1){
      uint32_t complete_flag = 0;
      uint32_t restart_timer_ms;

      restart_timer_ms = 0;

      prepare_to_wait(&cxp_device->cxp_transaction.wait_queue_head, &wait, TASK_INTERRUPTIBLE);

      if (cxp_device->cxp_transaction.cxp_transaction_state == cxp_idle){
         cxp_device->cxp_transaction.cxp_transaction_state = cxp_started;
         aslcxp_write32(cxp_device, FBD_CXP_ADDR_TO_DEVICE_COMMAND, cxp_command | cxp_byte_count);
         if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: wrote 0x%08X to command\n", __FUNCTION__, cxp_command | cxp_byte_count);
         spin_unlock_irqrestore(&cxp_device->interrupt_spinlock, irq_state);
         locked = 0;
         aslcxp_start_ack_timer(&cxp_device->cxp_transaction.ack_timer, 250);
      }

      if (signal_pending(current)){
          if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: Signal pending\n", __FUNCTION__);
          error = fatal_signal_pending(current) ? -ENODATA : -EAGAIN;
          complete_flag = 1;
      }

      spin_lock_irqsave(&cxp_device->interrupt_spinlock, irq_state);
      locked = 1;
      if (error){
         if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: setting complete flag for error %d\n", __FUNCTION__, error);
         complete_flag = 1;
      }
      else{
         switch(cxp_device->cxp_transaction.cxp_transaction_state){
            case cxp_started:
               break;
            case cxp_tentative_ack:
            case cxp_wait_ack:
               if (cxp_device->cxp_transaction.ack_timeout_ms){
                  restart_timer_ms = cxp_device->cxp_transaction.ack_timeout_ms;
                  cxp_device->cxp_transaction.ack_timeout_ms = 0;
                  aslcxp_restart_ack_timer(&cxp_device->cxp_transaction.ack_timer, restart_timer_ms);
               }
               break;
            case cxp_complete:
            case cxp_error:
               complete_flag = 1;
               break;
            case cxp_idle:
               if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: cxp_idle\n", __FUNCTION__);
               break;
            default:
               dev_err(cxp_device->device, "%s: Invalid state\n", __FUNCTION__);
               cxp_device->cxp_transaction.error_code = -EPROTO;
               cxp_device->cxp_transaction.protocol_error_code = 0;
               complete_flag = 1;
               break;
         }
      }
      if (complete_flag){
         cxp_device->cxp_transaction.cxp_transaction_state = cxp_idle;
         if (timeout_ok && (cxp_device->cxp_transaction.error_code == -ETIME) && (cxp_device->cxp_transaction.protocol_error_code == FBD_CXP_ASL_VALUE_TO_HOST_ERROR_ACK_TIMEOUT)){
            cxp_device->cxp_transaction.error_code = 0;
            cxp_device->cxp_transaction.protocol_error_code = 0;
            if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: Reset operation timeout allowed\n", __FUNCTION__);
         }
      }
      spin_unlock_irqrestore(&cxp_device->interrupt_spinlock, irq_state);
      locked = 0;

      if (aslcxp_hw_debug && restart_timer_ms) dev_printk(KERN_INFO, cxp_device->device, "%s: ack timer = %u (0x%08X)\n", __FUNCTION__, restart_timer_ms, restart_timer_ms);

      if (complete_flag){
         if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: stopping ack timer\n", __FUNCTION__);
          aslcxp_stop_ack_timer(&cxp_device->cxp_transaction.ack_timer);
          break;
      }

      schedule();
    }
    finish_wait(&cxp_device->cxp_transaction.wait_queue_head, &wait);

    if (error){
      dev_err(cxp_device->device, "%s: wait failed with %d", __FUNCTION__, error);
      goto bail;
    }

    error = cxp_device->cxp_transaction.error_code;

bail:
   if (locked) spin_unlock_irqrestore(&cxp_device->interrupt_spinlock, irq_state);
   if (aslcxp_hw_debug) dev_printk(KERN_INFO, cxp_device->device, "%s: return with %u (0x%08X)\n", __FUNCTION__, error, error);
   if (error == -EIO || error == -ETIME){
      cxp_device->cxp_last_protocol_error = cxp_device->cxp_transaction.protocol_error_code;
      dev_err(cxp_device->device, "%s: wait failed with EIO (%d) protocol error 0x%02X\n", __FUNCTION__, error, cxp_device->cxp_last_protocol_error);
   }
   else if (error){
      dev_err(cxp_device->device, "%s: wait failed with %d", __FUNCTION__, error);
   }
   return error;
}

uint32_t aslcxp_do_read32(struct aslcxp_device *cxp_device, uint32_t offset, const char *szWhence)
{
   uint32_t value = mda_read32(cxp_device->mdadev, offset);
   if (aslcxp_hw_debug > 1){
      dev_printk(KERN_INFO, cxp_device->device, "RD %08d %08X %08X %s\n", atomic_inc_return(&cxp_device->log_count), offset, value, szWhence);
   }
   return value;
}

void aslcxp_do_write32(struct aslcxp_device *cxp_device, uint32_t offset, uint32_t value, const char *szWhence)
{
   if (aslcxp_hw_debug > 1){
      dev_printk(KERN_INFO, cxp_device->device, "WR %08d %08X %08X %s\n", atomic_inc_return(&cxp_device->log_count), offset, value, szWhence);
   }
   mda_write32(cxp_device->mdadev, offset, value);
   return;
}

EXPORT_SYMBOL(aslcxp_link_is_active);

