#include "precomp.h"

#if !defined list_last_entry
#define list_last_entry(ptr, type, member) \
	list_entry((ptr)->prev, type, member)
#endif

extern int asldma_enable_dma_sync;

int asldma_debug_dma = 0;
int asldma_debug_dma_detail = 0;
int asldma_debug_event = 0;


int asldma_open(struct inode * inode, struct file * file)
{
   struct asldma_device *dma_device = NULL;
   int error = -EIO;

   dma_device = container_of(inode->i_cdev, struct asldma_device, cdev);

   spin_lock(&dma_device->open_lock);
   if ((file->f_flags & O_EXCL) && (dma_device->open_count > 0)){
      error = -EBUSY;
   }
   else{
      dma_device->open_count++;
      error = 0;
   }
   spin_unlock(&dma_device->open_lock);

   if (error == 0) file->private_data = dma_device;

   return error;
}

static void asldma_do_clear_event_queue(struct asldma_device *dma_device)
{
   unsigned long irq_state = 0;
   unsigned int event_queue_count = 0;
   struct asldma_event *event = NULL;
   struct asldma_event *tmp = NULL;

   spin_lock_irqsave(&dma_device->event_list_lock, irq_state);
   list_for_each_entry_safe(event, tmp, &dma_device->event_pending_list, this_event){
      list_del(&event->this_event);
      list_add_tail(&event->this_event, &dma_device->event_free_list);
      event_queue_count++;
   }
   MdaResetSequenceNumber(&dma_device->event_sequence_number);
   spin_unlock_irqrestore(&dma_device->event_list_lock, irq_state);

   if (event_queue_count > 0) dev_printk(KERN_INFO,
                              dma_device->device,
                              "%s: removed %u entries\n",
                              __FUNCTION__,
                              event_queue_count);

   return;
}

static void asldma_do_memory_unlock(struct asldma_device *dma_device, struct mda_dma_buffer *dma_buffer)
{
   if (asldma_debug_dma) dev_printk(KERN_INFO,
                                    dma_device->device,
                                    "%s: p=0x%016llX, l=0x%016llX => %d pages\n",
                                   __FUNCTION__,
                                   dma_buffer->user_addr_in,
                                   dma_buffer->user_size_in,
                                   dma_buffer->nPages);
   mda_dma_unlock(dma_device->mdadev, dma_buffer);
   kfree(dma_buffer->scatterlist);
   kfree(dma_buffer->page_ptrs);
   kfree(dma_buffer);
   return;
}

static void asldma_release_all_dma_buffers(struct asldma_device *dma_device)
{
   struct mda_dma_buffer *dma_buffer = NULL;
   struct mda_dma_buffer *tmp_buffer = NULL;

   spin_lock(&dma_device->dma_buffer_list_lock);
   list_for_each_entry_safe(dma_buffer, tmp_buffer, &dma_device->dma_buffer_list, this_dma_buffer){
      list_del(&dma_buffer->this_dma_buffer);
      if (asldma_debug_dma) dev_printk(KERN_INFO,
                                       dma_device->device,
                                       "%s: p=0x%016llX, l=0x%016llX => %d pages\n",
                                       __FUNCTION__,
                                       dma_buffer->user_addr_in,
                                       dma_buffer->user_size_in,
                                       dma_buffer->nPages);
      asldma_do_memory_unlock(dma_device, dma_buffer);
   }
   spin_unlock(&dma_device->dma_buffer_list_lock);

   return;
}

static void asldma_do_abort_event_thread(struct asldma_device *dma_device)
{
   unsigned long irq_state = 0;

   spin_lock_irqsave(&dma_device->event_list_lock, irq_state);
   dma_device->event_abort_reason = MDADMA_EVENT_ABORT_REASON_CLOSED;
   wake_up( &dma_device->event_wait_queue_head );
   spin_unlock_irqrestore(&dma_device->event_list_lock, irq_state);

   return;
}

static void asldma_finish(struct asldma_device *dma_device)
{
   asldma_stop_risc_engine(dma_device);  
   asldma_do_abort_event_thread(dma_device);
   asldma_do_clear_event_queue(dma_device);
   asldma_release_all_dma_buffers(dma_device);
   if (asldma_debug_dma_detail) dev_printk(KERN_INFO, dma_device->device, "%s: complete\n", __FUNCTION__);

   return;
}

int asldma_release(struct inode * inode, struct file * file)
{
   struct asldma_device *dma_device = NULL;
   int error = -EIO;
   int fDmaStop = 0;

   dma_device = container_of(inode->i_cdev, struct asldma_device, cdev);

   spin_lock(&dma_device->open_lock);
   if (dma_device->open_count > 0){
      dma_device->open_count--;
      if ((dma_device->open_count == 0) || (file == dma_device->file_write_exclusive)){
         dma_device->file_write_exclusive = NULL;
		 fDmaStop =1;
      }
      error = 0;
   }
   else{
      error = -ENOENT;
   }
   spin_unlock(&dma_device->open_lock);

   if (error == -ENOENT) dev_err(dma_device->device, "%s: Device not open\n", __FUNCTION__);

   if (error == 0){
      if (fDmaStop) asldma_finish(dma_device);
      file->private_data = NULL;
   }

   return error;
}

static int asldma_write_exclusive_acquire(struct asldma_device *dma_device, struct file * file)
{
   int error = 0;

   spin_lock(&dma_device->open_lock);
   if (dma_device->open_count < 1) error = -ENXIO;
   else if (dma_device->file_write_exclusive == file) error = -EIO;
   else if (dma_device->file_write_exclusive != NULL) error = -EBUSY;
   if (!error){
      unsigned long irq_state = 0;
      dma_device->file_write_exclusive = file;
      spin_lock_irqsave(&dma_device->event_list_lock, irq_state);
      dma_device->event_abort_reason = 0;
      MdaResetSequenceNumber(&dma_device->event_sequence_number);
      spin_unlock_irqrestore(&dma_device->event_list_lock, irq_state);
   }
   spin_unlock(&dma_device->open_lock);
   return error;
}

static int asldma_write_exclusive_release(struct asldma_device *dma_device, struct file * file)
{
   int error = 0;

   spin_lock(&dma_device->open_lock);
   if (dma_device->open_count < 1) error = -ENXIO;
   else if (dma_device->file_write_exclusive != file) error = -EIO;
   if (!error) dma_device->file_write_exclusive = NULL;
   spin_unlock(&dma_device->open_lock);
   return error;
}

static int asldma_check_exclusive(struct asldma_device *dma_device, struct file * file)
{
   int error = 0;
   spin_lock(&dma_device->open_lock);
   if (dma_device->file_write_exclusive != file) error = -EBUSY;
   spin_unlock(&dma_device->open_lock);
   return error;
}

static long asldma_ioctl_reglist (struct asldma_device *dma_device, struct file * file, mda_io_buffer *user_io)
{
   int error = asldma_check_exclusive(dma_device, file);
   int writeok = (error) ? 0 : 1;
   return mda_ioctl_reglist(dma_device->mdadev, user_io, writeok);
}

static long asldma_check_permission(struct file * file, unsigned int cmd)
{
   struct asldma_device * dma_device = file->private_data;
   long error = -EBUSY;

   switch(cmd){
      case CDA_IOCTL_REGLIST:
      case MDA_IOCTL_SYSTEM_REGLIST:
      case CDA_IOCTL_CLAIM:
      case CDA_IOCTL_DEVICEINFO:
      case MDA_IOCTL_DMA_INFO:
      case MDA_IOCTL_GET_TIME:
      case MDA_IOCTL_GET_VERSION:
      case MDA_IOCTL_GET_UNIQUE_ID:
         error = 0;
         break;
      default:
         error = asldma_check_exclusive(dma_device, file);
         break;
   }
   return error;
}

static volatile char clr_mda_c;
static volatile char clr_mda_d;

int asldma_check_copy_to_user(void *to, void*from, size_t sizeCopy)
{
   size_t index = 0;
   char *pFrom = (char*)from;
   for (index = 0; index < sizeCopy; index++) clr_mda_c = pFrom[index];
   return copy_to_user(to, from, sizeCopy);
}

static long asldma_mda_event(struct asldma_device *dma_device, mda_io_buffer *user_io)
{
   long error = -EIO;
   MDA_IoctlEvent mda_event = {0};
   DEFINE_WAIT(wait);

   if (user_io->output_buffer.user_address == NULL){
      error = -EINVAL;
      dev_err(dma_device->device, "%s: Output buffer is NULL\n", __FUNCTION__);
      goto bail;
   }
   if (user_io->output_buffer.user_size < sizeof(MDA_IoctlEvent)){
      dev_err(dma_device->device, "%s: Output buffer is too small\n", __FUNCTION__);
      error = -EFAULT;
      goto bail;
   }
   error = asldma_check_copy_to_user(user_io->output_buffer.user_address, &mda_event, sizeof(mda_event));
   if (error){
      dev_err(dma_device->device, "%s: asldrv_copy_to_user returned %d\n", __FUNCTION__, (int)error);
      goto bail;
   }
   user_io->output_buffer.user_size = 0;

   error = -EIO;

   while(1){
      unsigned long irq_state = 0;
      uint32_t event_queue_overflow_count = 0;
      uint32_t event_fifo_overflow_count = 0;

      prepare_to_wait(&dma_device->event_wait_queue_head, &wait, TASK_INTERRUPTIBLE);
      spin_lock_irqsave(&dma_device->event_list_lock, irq_state);
      if (dma_device->event_abort_reason){
         error = -ENXIO;
      }
      else if (!list_empty(&dma_device->event_pending_list)){
         struct asldma_event *event = NULL;

         event = list_first_entry(&dma_device->event_pending_list, struct asldma_event, this_event);
         memcpy(&mda_event, &event->event, sizeof(mda_event));
         list_del(&event->this_event);
         list_add_tail(&event->this_event, &dma_device->event_free_list);
         event_queue_overflow_count = dma_device->event_queue_overflow_count;
         dma_device->event_queue_overflow_count = 0;
         event_fifo_overflow_count = dma_device->event_fifo_overflow_count;
         dma_device->event_fifo_overflow_count = 0;
         error = 0;
      }
      spin_unlock_irqrestore(&dma_device->event_list_lock, irq_state);
      
      if (error == -ENXIO){
         if (asldma_debug_event) dev_err(dma_device->device, "%s: Aborted\n", __FUNCTION__);
         break;
      }

      if (!error){
         mda_event.qwDispatchTimestamp = mda_gettime();
         if (event_queue_overflow_count){
            mda_event.qwFlags |= MDA_IOCTL_MDA_FLAG_INTERRUPT_QUEUE_OVERRUN;
            dev_err(dma_device->device, "%s: Interrupt queue overflowed %u times\n", __FUNCTION__, event_queue_overflow_count);
         }
         if (event_fifo_overflow_count){
            mda_event.qwFlags |= MDA_IOCTL_MDA_FLAG_INTERRUPT_FIFO_OVERRUN;
            dev_err(dma_device->device, "%s: Interrupt FIFO overflowed %u times\n", __FUNCTION__, event_fifo_overflow_count);
         }
         break;
      }

      if (signal_pending(current)){
         error = -EAGAIN;
         break;
      }
      schedule();
   }
   finish_wait(&dma_device->event_wait_queue_head, &wait);
   
bail:
   if (!error){
      error = copy_to_user(user_io->output_buffer.user_address, &mda_event, sizeof(mda_event));
      if (!error) user_io->output_buffer.user_size = sizeof(mda_event);
   }

   if (asldma_debug_event && !error){
      dev_printk(KERN_INFO,
                 dma_device->device,
                 "%s: %d\n",
                 __FUNCTION__,
                mda_event.sEvent.dwEvent);
   }
   if (error && (error != -EAGAIN) && (error != -ENXIO)){
      dev_err(dma_device->device,
      "%s returning %d\n",
      __FUNCTION__,
      (int)error);
   }

   return error;
}

static long asldma_event_exit(struct asldma_device *dma_device, mda_io_buffer *user_io)
{
   if (asldma_debug_event) dev_printk(KERN_INFO, dma_device->device, "%s\n", __FUNCTION__);
   asldma_do_clear_event_queue(dma_device);
   asldma_do_abort_event_thread(dma_device);
   return 0;
}

static long asldma_event_put(struct asldma_device *dma_device, mda_io_buffer *user_io)
{
   int error = -EIO;
   unsigned long irq_state = 0;
   struct asldma_event *queued_event = NULL;
   MDA_IoctlEvent user_mda_event = {0};

   if (user_io->param_buffer.user_address == NULL){
      error = -EINVAL;
      dev_err(dma_device->device, "%s: Parameter buffer is NULL\n", __FUNCTION__);
      goto bail;
   }
   if (user_io->param_buffer.user_size < sizeof(user_mda_event)){
      dev_err(dma_device->device, "%s: Parameter buffer is too small\n", __FUNCTION__);
      error = -EFAULT;
      goto bail;
   }
   error = copy_from_user(&user_mda_event, user_io->param_buffer.user_address, sizeof(user_mda_event));
   if (error){
      dev_err(dma_device->device, "%s: copy from use returned %d\n", __FUNCTION__, error);
      goto bail;
   }

   spin_lock_irqsave(&dma_device->event_list_lock, irq_state);
   user_mda_event.qwSequenceNumber = MdaGetSequenceNumber(&dma_device->event_sequence_number);
   if (!list_empty(&dma_device->event_free_list)){
      queued_event = list_first_entry(&dma_device->event_free_list, struct asldma_event, this_event);
      list_del(&queued_event->this_event);
      list_add_tail(&queued_event->this_event, &dma_device->event_pending_list);
      error = 0;
   }
   else{ /* Queue overrun  - must overwrite the last pending entry */
      queued_event = list_last_entry(&dma_device->event_pending_list, struct asldma_event, this_event);
      dma_device->event_queue_overflow_count++;
      error = -ENOMEM;
   }
   memcpy(&queued_event->event, &user_mda_event, sizeof(user_mda_event));
   wake_up( &dma_device->event_wait_queue_head );
   spin_unlock_irqrestore(&dma_device->event_list_lock, irq_state);

bail:
   return error;
}

static long asldma_clear_event_queue(struct asldma_device *dma_device, mda_io_buffer *user_io)
{
   asldma_do_clear_event_queue(dma_device);
   return 0;
}

static long asldma_dma_info(struct asldma_device *dma_device, mda_io_buffer *user_io)
{
   long error = -EIO;
   MDA_DmaInfo dma_info = {0};

   if (user_io->output_buffer.user_address == NULL){
      error = -EINVAL;
      dev_err(dma_device->device, "%s: Output buffer is NULL\n", __FUNCTION__);
      goto bail;
   }
   if (user_io->output_buffer.user_size < sizeof(MDA_DmaInfo)){
      dev_err(dma_device->device, "%s: Output buffer is too small\n", __FUNCTION__);
      error = -EFAULT;
      goto bail;
   }
   user_io->output_buffer.user_size = sizeof(MDA_DmaInfo);
   dma_info.qwMaxMdl = MDA_DMAINFO_UNLIMITED;
   dma_info.qwMaxDmaTotal = MDA_DMAINFO_UNLIMITED;
   dma_info.qwMaxReadFragment = MDA_DMAINFO_UNLIMITED;
   dma_info.qwMaxWriteFragment = MDA_DMAINFO_UNLIMITED;
   dma_info.qwMaxTransfer = MDA_DMAINFO_UNLIMITED;
   dma_info.qwMaxSGElements = MDA_DMAINFO_UNLIMITED/PAGE_SIZE;
   error = copy_to_user(user_io->output_buffer.user_address, &dma_info, sizeof(dma_info));
bail:
   return error;
}

static long asldma_gettime(struct asldma_device *dma_device, mda_io_buffer *user_io)
{
   long error = -EIO;
   uint64_t qwTime = mda_gettime();

   if (user_io->output_buffer.user_address == NULL){
      error = -EINVAL;
      dev_err(dma_device->device, "%s: Output buffer is NULL\n", __FUNCTION__);
      goto bail;
   }
   if (user_io->output_buffer.user_size < sizeof(qwTime)){
      dev_err(dma_device->device, "%s: Output buffer is too small\n", __FUNCTION__);
      error = -EFAULT;
      goto bail;
   }
   user_io->output_buffer.user_size = sizeof(qwTime);
   error = copy_to_user(user_io->output_buffer.user_address, &qwTime, sizeof(qwTime));
bail:
   return error;
}

static void asldma_build_dma_sg_user_table(struct mda_dma_buffer *dma_buffer, MDA_SgEntry *sgout)
{
   int i;
   struct scatterlist *sglist, *sg;
   dma_addr_t entry_address;
   unsigned int offset, entry_length, len;
   u64 total, page_index;

   total = 0;
   page_index = 0;
   offset = (unsigned int)(dma_buffer->user_addr_in - dma_buffer->user_base);
   sglist = dma_buffer->scatterlist;

   for_each_sg(sglist, sg, dma_buffer->nSgElements, i){
      entry_address = sg_dma_address(sg);
      entry_length = sg_dma_len(sg);
      while(offset < entry_length && page_index < dma_buffer->nPages && total < dma_buffer->user_size_in){
         sgout[page_index].uiPhysAddr = entry_address + offset;
         if((offset + PAGE_SIZE) >= entry_length){
            len = entry_length - offset;
         } else {
            len = PAGE_SIZE;
            if(!page_index) 
               len -= offset;
         }
         if(page_index == (dma_buffer->nPages - 1))
            len = dma_buffer->user_size_in - total;
         sgout[page_index].uiPhysLen = len;
         offset += len;
         total += len;
         page_index++;			
      }
      offset = 0;
   }
}

static long asldma_dma_copyout(struct asldma_device *dma_device, struct mda_dma_buffer *dma_buffer, mda_io_buffer *user_io)
{
	long error = -EIO;
	int entry_index;
	MDA_SgEntry * sgout = NULL;
	size_t size_out = dma_buffer->nPages * sizeof(*sgout);

   sgout = kzalloc(size_out, GFP_KERNEL);
   if (sgout == NULL){
      dev_err(dma_device->device, "%s: failed to allocate temporary buffer\n", __FUNCTION__);
      error = -ENOMEM;
      goto bail;
   }

   asldma_build_dma_sg_user_table(dma_buffer, sgout);

   error = copy_to_user(user_io->output_buffer.user_address, sgout, size_out);
   user_io->output_buffer.user_size = size_out;

   if (asldma_debug_dma){
      dev_printk(KERN_INFO,
                 dma_device->device,
                 "%s: p=0x%016llX, l=0x%016llX => %d pages, %lld bytes\n",
                 __FUNCTION__,
                 dma_buffer->user_addr_in,
                 dma_buffer->user_size_in,
                 dma_buffer->nPages,
                 (long long int)size_out);
   }
   if (asldma_debug_dma_detail){
      for(entry_index = 0; entry_index < dma_buffer->nPages; entry_index++){
         dev_printk(KERN_INFO,
                    dma_device->device,
                    "%s: %d %016llX/%016llX\n",
                    __FUNCTION__,
                    entry_index,
                    sgout[entry_index].uiPhysAddr,
                    sgout[entry_index].uiPhysLen);
      }
   }
bail:
   if (sgout) kfree(sgout);
   return error;
}

static void asldma_dma_destroy_callback(void *data)
{
   struct asldma_device *dma_device = (struct asldma_device*)data;

   dev_printk(KERN_INFO, dma_device->device, "%s: stopping RISC engine\n", __FUNCTION__);
   asldma_stop_risc_engine(dma_device);
   return;
}

static long asldma_memory_lock(struct asldma_device *dma_device, mda_io_buffer *user_io, enum dma_data_direction direction, enum asldma_memory_type memory_type)
{
   long error = -EIO;
   struct mda_dma_buffer *dma_buffer = NULL;
   MDA_SgHeader mda_header = {0};
   u64 first_page_offset = 0;
   u64 total_map_bytes = 0;
   const u64 page_size = (u64)PAGE_SIZE;

   if (user_io->param_buffer.user_address == NULL){
      dev_err(dma_device->device, "%s: NULL SG Header\n", __FUNCTION__);
      error = -EINVAL;
      goto err_header;
   }
   if (user_io->param_buffer.user_size < sizeof(MDA_SgHeader)){
      dev_err(dma_device->device, "%s: Input buffer too small\n", __FUNCTION__);
      error = -EFAULT;
      goto err_header;
   }
   error = copy_from_user(&mda_header, user_io->param_buffer.user_address, sizeof(mda_header));
   if (error){
      dev_err(dma_device->device, "%s: copy_from_user returned %d\n", __FUNCTION__, (int)error);
      goto err_header;
   }

   dma_buffer = kzalloc(sizeof(*dma_buffer), GFP_KERNEL);
   if (!dma_buffer){
      dev_err(dma_device->device, "%s: Failed to allocate DMA buffer structure\n", __FUNCTION__);
      error = -ENOMEM;
      goto err_buffer;
   }

   INIT_LIST_HEAD(&dma_buffer->this_dma_buffer);

   dma_buffer->user_addr_in = _MDA_SGTable_Get_VirtualAddr(&mda_header) +
                              _MDA_SGTable_Get_StartOffset(&mda_header);
   dma_buffer->user_size_in = _MDA_SGTable_Get_TransferSize(&mda_header);
   dma_buffer->direction = direction;
   dma_buffer->memory_type = memory_type;
   dma_buffer->destroy_callback = asldma_dma_destroy_callback;
   dma_buffer->destroy_data = dma_device;
   first_page_offset = dma_buffer->user_addr_in % page_size;
   dma_buffer->user_base = dma_buffer->user_addr_in - first_page_offset;
   total_map_bytes = dma_buffer->user_size_in + first_page_offset;
   dma_buffer->nPages = (int)(total_map_bytes / page_size);
   if (total_map_bytes % page_size) dma_buffer->nPages++;
   dma_buffer->nSgElements = 0;

   if (user_io->output_buffer.user_address == NULL){
      dev_err(dma_device->device, "%s: NULL SG table\n", __FUNCTION__);
      error = -EINVAL;
      goto err_output;
   }
   if (user_io->output_buffer.user_size < (dma_buffer->nPages * sizeof(MDA_SgEntry))){
      dev_err(dma_device->device, "%s: Output buffer too small\n", __FUNCTION__);
      error = -EFAULT;
      goto err_output;
   }

   dma_buffer->page_ptrs = kzalloc(dma_buffer->nPages * sizeof(struct page *), GFP_KERNEL);
   if (!dma_buffer->page_ptrs){
      dev_err(dma_device->device, "%s: Failed to allocate pager structures\n", __FUNCTION__);
      error = -ENOMEM;
      goto err_pages;
   }

   if (asldma_debug_dma_detail){
      dev_printk(KERN_INFO,
                 dma_device->device,
                 "%s: Allocating %lu bytescatterlist for %lu pages\n",
                  __FUNCTION__,
                  (long unsigned int)dma_buffer->nPages * sizeof(struct scatterlist),
                  (long unsigned int)dma_buffer->nPages);
   }

   dma_buffer->scatterlist = kzalloc(dma_buffer->nPages * sizeof(struct scatterlist), GFP_KERNEL);
   if (!dma_buffer->scatterlist){
      dev_err(dma_device->device, "%s: Failed to allocate page structures\n", __FUNCTION__);
      error = -ENOMEM;
      goto err_scatter;
   }

   if (asldma_debug_dma_detail){
      dev_printk(KERN_INFO,
                 dma_device->device,
                 "%s: user %p/%p, base %p, %d pages, %s, %s\n",
                  __FUNCTION__,
                  (void*)dma_buffer->user_addr_in,
                  (void*)dma_buffer->user_size_in,
                  (void*)dma_buffer->user_base,
                  dma_buffer->nPages,
                  dma_buffer->direction == DMA_FROM_DEVICE ? "from device" : "to device",
                  dma_buffer->memory_type == system_memory ? "system memory" : "nvidia p2p memory");
   }
   error = mda_dma_lock(dma_device->mdadev, dma_buffer);
   if (error){
      dev_err(dma_device->device, "%s: mda_dma_lock returned %d\n", __FUNCTION__, (int)error);
      goto err_lock;
   }

   error = asldma_dma_copyout(dma_device, dma_buffer, user_io);
   if (error){
      dev_err(dma_device->device, "%s: asldma_dma_copyout returned %d\n", __FUNCTION__, (int)error);
      goto err_copyout;
   }
   spin_lock(&dma_device->dma_buffer_list_lock);
   list_add_tail(&dma_buffer->this_dma_buffer, &dma_device->dma_buffer_list);
   spin_unlock(&dma_device->dma_buffer_list_lock);

   goto finished;

err_copyout:
   mda_dma_unlock(dma_device->mdadev, dma_buffer);

err_lock:
   kfree(dma_buffer->scatterlist);

err_scatter:
   kfree(dma_buffer->page_ptrs);

err_pages:
err_output:
   kfree(dma_buffer);

err_buffer:
err_header:

finished:
   return error;
}

static long asldma_memory_unlock(struct asldma_device *dma_device, mda_io_buffer *user_io)
{
   long error = -EIO;
   MDA_SgHeader mda_header = {0};
   struct mda_dma_buffer *dma_buffer = NULL;
   struct mda_dma_buffer *tmp_buffer = NULL;
   struct mda_dma_buffer *dma_buffer_to_free = NULL;
   u64 user_addr_in = 0;

   if (user_io->param_buffer.user_address == NULL){
      dev_err(dma_device->device, "%s: NULL SG Header\n", __FUNCTION__);
      error = -EINVAL;
      goto bail;
   }
   if (user_io->param_buffer.user_size < sizeof(MDA_SgHeader)){
      dev_err(dma_device->device, "%s: Input buffer too small\n", __FUNCTION__);
      error = -EFAULT;
      goto bail;
   }
   error = copy_from_user(&mda_header, user_io->param_buffer.user_address, sizeof(mda_header));
   if (error){
      dev_err(dma_device->device, "%s: copy_from_user returned %d\n", __FUNCTION__, (int)error);
      goto bail;
   }

   user_addr_in = _MDA_SGTable_Get_VirtualAddr(&mda_header) +
                  _MDA_SGTable_Get_StartOffset(&mda_header);

   error = -EINVAL;
   spin_lock(&dma_device->dma_buffer_list_lock);
   list_for_each_entry_safe(dma_buffer, tmp_buffer, &dma_device->dma_buffer_list, this_dma_buffer){
      if (dma_buffer->user_addr_in == user_addr_in){
         list_del(&dma_buffer->this_dma_buffer);
         dma_buffer_to_free = dma_buffer;
         error = 0;
      }
   }
   spin_unlock(&dma_device->dma_buffer_list_lock);

   if (error){
      dev_err(dma_device->device, "%s: Buffer not locked\n", __FUNCTION__);
      goto bail;
   }

   asldma_do_memory_unlock(dma_device, dma_buffer_to_free);

bail:
   return error;
}

static long asldma_sync_dma_buffer_from_sg_user_header(struct asldma_device *dma_device, MDA_SgHeader *user_sg_header)
{
   struct mda_dma_buffer *dma_buffer = NULL;
   struct mda_dma_buffer *tmp_buffer = NULL;

   if (!asldma_enable_dma_sync) goto bail;

   spin_lock(&dma_device->dma_buffer_list_lock);
   list_for_each_entry_safe(dma_buffer, tmp_buffer, &dma_device->dma_buffer_list, this_dma_buffer){
      if(user_sg_header->uiVirtualAddr == dma_buffer->user_addr_in){
         struct mda_device *mda_device = dma_device->mdadev;
         struct pci_dev *pci_device = (struct pci_dev *)mda_device->parent->parent->parent->pcidev;
         struct scatterlist *sg_list = (struct scatterlist *)dma_buffer->scatterlist;

         if(dma_buffer->direction == DMA_TO_DEVICE){
            /* Flushing RISC buffer from CPU to FB */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
            dma_sync_sg_for_device(&pci_device->dev, sg_list, dma_buffer->nPages, DMA_TO_DEVICE);
#else
            pci_dma_sync_sg_for_device(pci_device, sg_list, dma_buffer->nPages, DMA_TO_DEVICE);
#endif
		 } else {
            /* Flushing image capture from FB to CPU */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
            dma_sync_sg_for_cpu(&pci_device->dev, sg_list, dma_buffer->nPages, DMA_FROM_DEVICE);
#else
            pci_dma_sync_sg_for_cpu(pci_device, sg_list, dma_buffer->nPages, DMA_FROM_DEVICE);
#endif
		 }
      }
   }
   spin_unlock(&dma_device->dma_buffer_list_lock);

bail:
   return 0;
}

static long asldma_memory_flush(struct asldma_device *dma_device, mda_io_buffer *user_io)
{
   long error = 0;

   if(user_io!=NULL){
      if(user_io->output_buffer.user_address!=NULL && user_io->output_buffer.user_size==sizeof(MDA_SgHeader)){
         MDA_SgHeader *userSgHeader = (MDA_SgHeader *)user_io->output_buffer.user_address;
         error = asldma_sync_dma_buffer_from_sg_user_header(dma_device, userSgHeader);
         user_io->output_buffer.user_size = 0;
      }
   }
   return error;
}

static long asldma_ioctl_process(struct file * file, unsigned int cmd, mda_io_buffer *user_io)
{
   long error = 0;
   struct asldma_device * dma_device = file->private_data;

   error = asldma_check_permission(file, cmd);
   if (error){
      dev_err(dma_device->device, "%s: Operation 0x%08X required exclusive access\n", __FUNCTION__, cmd);
      goto bail;
   }

   switch(cmd){
   case CDA_IOCTL_REGLIST:
   case MDA_IOCTL_SYSTEM_REGLIST:
      error = asldma_ioctl_reglist(dma_device, file, user_io);
      break;

   case CDA_IOCTL_CLAIM:
      error = asldma_write_exclusive_acquire(dma_device, file);
      break;

   case MDA_IOCTL_RELEASE:
      error = asldma_write_exclusive_release(dma_device, file);
      break;

   case MDA_IOCTL_MDA_EVENT:
      error = asldma_mda_event(dma_device, user_io);
      break;

   case CDA_IOCTL_EVENT_PUT:
      error = asldma_event_put(dma_device, user_io);
      break;

   case MDA_IOCTL_EVENT_EXIT:
      error = asldma_event_exit(dma_device, user_io);
      break;

   case MDA_IOCTL_CLEAR_EVENT_QUEUE:
      error = asldma_clear_event_queue(dma_device, user_io);
      break;

   case MDA_IOCTL_DMA_INFO:
      error = asldma_dma_info(dma_device, user_io);
      break;

   case MDA_IOCTL_MEMORY_LOCK_IMAGE:
      error = asldma_memory_lock(dma_device, user_io, DMA_FROM_DEVICE, system_memory);
      break;

   case MDA_IOCTL_MEMORY_LOCK_RISC:
      error = asldma_memory_lock(dma_device, user_io, DMA_TO_DEVICE, system_memory);
      break;

   case MDA_IOCTL_MEMORY_LOCK_CUDA:
      error = asldma_memory_lock(dma_device, user_io, DMA_FROM_DEVICE, nvidia_p2p_memory);
      break;

   case MDA_IOCTL_MEMORY_UNLOCK:
      error = asldma_memory_unlock(dma_device, user_io);
      break;

   case MDA_IOCTL_MEMORY_FLUSH:
      error = asldma_memory_flush(dma_device, user_io);
      break;

   case MDA_IOCTL_EVENT_ENABLE:
      mda_fpga_is_reconfiguring(dma_device->mdadev, 0);
      error = 0;
      break;

   case MDA_IOCTL_EVENT_DISABLE:
      mda_fpga_is_reconfiguring(dma_device->mdadev, 1);
      error = 0;
      break;

   case MDA_IOCTL_CONFIG_RESTORE:
      error = (long)mda_config_space_restore(dma_device->mdadev);
      break;

   case MDA_IOCTL_GET_TIME:
      error = asldma_gettime(dma_device, user_io);
      break;

   case MDA_IOCTL_GET_VERSION:
      error = mda_ioctl_version(user_io, ASLDMA_VERSION);
      break;

   case MDA_IOCTL_GET_UNIQUE_ID:
      error = mda_ioctl_unique_id(user_io, dma_device->mdadev);
      break;

   default:
      error = -ENOTSUPP;
      break;
   }
bail:
   if (error && asldma_debug_dma){
      dev_err(dma_device->device,
              "%s: cmd = %x, returning %d\n",
              __FUNCTION__,
              cmd,
              (int)error);
   }
   return error;
}

long asldma_ioctl(struct file * file, unsigned int cmd, unsigned long arg)
{
   long error = -EINVAL;
   mda_io_buffer user64_io_buffer;

   memset(&user64_io_buffer, 0, sizeof(user64_io_buffer));

   if (arg != 0){
      error = copy_from_user(&user64_io_buffer, (void*)arg, sizeof(user64_io_buffer));
      if (error) goto bail;
      error = asldma_ioctl_process(file, cmd, &user64_io_buffer);
   }
bail:
   if (!error){
      error = copy_to_user((void*)arg, &user64_io_buffer, sizeof(user64_io_buffer));
   }
   return error;
}

#if defined __LP64__

long asldma_compat(struct file * file, unsigned int cmd, unsigned long arg)
{
   long error = -EINVAL;
   mda_io_buffer user64_io_buffer;
   struct mda_io32_buffer user32_io_buffer;

   memset(&user64_io_buffer, 0, sizeof(user64_io_buffer));
   memset(&user32_io_buffer, 0, sizeof(user32_io_buffer));

   if (arg != 0){
      error = copy_from_user(&user32_io_buffer, (void*)arg, sizeof(user32_io_buffer));
      if (error) goto bail;
      user64_io_buffer.param_buffer.user_address = (void*)(u64)user32_io_buffer.param_buffer.user_address;
      user64_io_buffer.param_buffer.user_size = (size_t)user32_io_buffer.param_buffer.user_size;
      user64_io_buffer.input_buffer.user_address = (void*)(u64)user32_io_buffer.input_buffer.user_address;
      user64_io_buffer.input_buffer.user_size = (size_t)user32_io_buffer.input_buffer.user_size;
      user64_io_buffer.output_buffer.user_address = (void*)(u64)user32_io_buffer.output_buffer.user_address;
      user64_io_buffer.output_buffer.user_size = (size_t)user32_io_buffer.output_buffer.user_size;
      error = asldma_ioctl_process(file, cmd, &user64_io_buffer);
   }
bail:
   if (!error){
      user32_io_buffer.output_buffer.user_size = (uint32_t) user64_io_buffer.output_buffer.user_size;
      error = copy_to_user((void*)arg, &user32_io_buffer, sizeof(user32_io_buffer));
   }
   return error;
}

#endif

struct file_operations asldma_ops = {
   .owner = THIS_MODULE,
   .open  = asldma_open,
   .release = asldma_release,
   .unlocked_ioctl = asldma_ioctl,
#if defined __LP64__
   .compat_ioctl = asldma_compat,
#endif
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
MODULE_DEVICE_TABLE(file_operations, asldma_ops);
#else
MODULE_GENERIC_TABLE(file_operations, asldma_ops);
#endif


