#include "precomp.h"

#include "../../all/fbd_acq.h"
#include "../../all/phd_drv.h"

static void asldma_isr_post_event(struct asldma_device *dma_device, uint32_t dwEvent, uint32_t dwData)
{
   spin_lock(&dma_device->event_list_lock);
   if (!list_empty(&dma_device->event_free_list)){
      struct asldma_event * event;
      event = list_first_entry(&dma_device->event_free_list, struct asldma_event, this_event);
      list_del(&event->this_event);
      event->event.sEvent.dwEvent = dwEvent;
      event->event.sEvent.dwData = dwData;
      event->event.qwSequenceNumber = MdaGetSequenceNumber(&dma_device->event_sequence_number);
      event->event.qwFlags = 0;
      event->event.qwInterruptTimestamp = mda_gettime();
      event->event.qwDispatchTimestamp = 0;
      event->event.qwSize = sizeof(event->event);
      list_add_tail(&event->this_event, &dma_device->event_pending_list);
   }
   else{
      dma_device->event_queue_overflow_count++;
      dev_err(dma_device->device, "%s: Queue overflow\n", __FUNCTION__);
   }
   spin_unlock(&dma_device->event_list_lock);
}

static irqreturn_t asldma_firebird_isr(struct mda_device *mda_device, struct asldma_device *dma_device)
{
   irqreturn_t irqreturn = IRQ_NONE;
   uint32_t interrupt_fifo_level = 0;
   uint32_t interrupt_status = 0;

   interrupt_fifo_level = mda_read32(mda_device, FBD_ADDR_STAT_INTRPT_FIFO_STATUS);
   if (interrupt_fifo_level & FBD_VALUE_STAT_INTRPT_FIFO_STATUS_OVERFLOW){
      mda_write32(mda_device, FBD_ADDR_STAT_INTRPT_FIFO_CONTROL, FBD_VALUE_STAT_INTRPT_FIFO_STATUS_OVERFLOW);
      irqreturn = IRQ_HANDLED;
      spin_lock(&dma_device->event_list_lock);
      dma_device->event_fifo_overflow_count++;
      spin_unlock(&dma_device->event_list_lock);
      dev_err(dma_device->device, "%s: FIFO overflow\n", __FUNCTION__);
   }
   interrupt_fifo_level &= FBD_VALUE_STAT_INTRPT_FIFO_STATUS_MASK;

   while(interrupt_fifo_level--){
      irqreturn = IRQ_HANDLED;
      interrupt_status = mda_read32(mda_device, FBD_ADDR_STAT_INTRPT_FIFO_EVENT);
      asldma_isr_post_event(dma_device, interrupt_status, 0);
   }

   if (irqreturn == IRQ_HANDLED) wake_up( &dma_device->event_wait_queue_head );

   return irqreturn;

}

static irqreturn_t asldma_phoenix_isr(struct mda_device *mda_device, struct asldma_device *dma_device)
{
   irqreturn_t irqreturn = IRQ_NONE;
   uint32_t status = 0;

   if (dma_device->fpga_reprogramming){
      goto bail;
   }

   /* Check if card is asserting INT */
   status = mda_read32(dma_device->mdadev, PHD_ADDR_STAT_INTRPT);

   if (status != 0){
      mda_write32(dma_device->mdadev, PHD_ADDR_REG_INTRPT_CLR, status ); /* Clear the IRQ */
      asldma_isr_post_event(dma_device, status, 0);
      irqreturn = IRQ_HANDLED;
    }
bail:
   return irqreturn;
}

irqreturn_t asldma_isr(struct mda_device *mda_device, void *data)
{
   irqreturn_t irqreturn = IRQ_NONE;
   struct asldma_device *dma_device = data;

   if (dma_device->baseboard_type == board_type_FireBird){
      irqreturn = asldma_firebird_isr(mda_device, dma_device);
   }
   else if (dma_device->baseboard_type == board_type_Phoenix){
      irqreturn = asldma_phoenix_isr(mda_device, dma_device);
   }
   return irqreturn;
}

static void asldma_firebird_stop_risc_engine(struct asldma_device *dma_device)
{
   mda_write32(dma_device->mdadev, FBD_ADDR_CTRL_RISC, 0);
   return;
}

static void asldma_phoenix_stop_risc_engine(struct asldma_device *dma_device)
{
   mda_write32(dma_device->mdadev, PHD_ADDR_REG_RISC_CTRL, 0);           /* Stop DMA channel A */
   mda_write32(dma_device->mdadev, PHD_ADDR_REG_RISC_CTRL + 0x100, 0);   /* Stop DMA channel B */
   mda_write32(dma_device->mdadev, PHD_ADDR_REG_INTRPT_CTRL, 0);         /* Disarm interrupts */
   mda_write32(dma_device->mdadev, PHD_ADDR_REG_INTRPT_CLR, 0xFFFFFFFF); /* Clear any pending  interrupts */
   return;
}

void asldma_stop_risc_engine(struct asldma_device *dma_device)
{
   if (dma_device->baseboard_type == board_type_FireBird) asldma_firebird_stop_risc_engine(dma_device);
   else if (dma_device->baseboard_type == board_type_Phoenix) asldma_phoenix_stop_risc_engine(dma_device);
   return;
}


