#include "precomp.h"
#include "boards.h"

u32 g_debug = 0;

void aslenum_write_exclusive_timeout(unsigned long callback_data);
long aslenum_ioctl_get_unique_id(mda_io_buffer *user_io, struct aslenum_device *pEnumDevice);

int aslenum_open(struct inode * inode, struct file * file)
{
   int error = -EIO;
   struct aslenum_device *baseboard = NULL;
   struct aslenum_file *pEnumFile = kzalloc(sizeof(*pEnumFile), GFP_KERNEL);

   if (!pEnumFile){
      error = -ENOMEM;
      goto bail;
   }

   baseboard = container_of(inode->i_cdev, struct aslenum_device, cdev);
   if (!baseboard){
      error = -ENXIO;
      goto bail;
   }

   if (g_debug) dev_printk(KERN_INFO, baseboard->device, "Entered aslenum_open %p %p\n", baseboard, file);

   pEnumFile->pEnumDevice = baseboard;
   pEnumFile->write_lock_timeout_ms = ASLENUM_WRITE_LOCK_TIMEOUT_DEFAULT;
   pEnumFile->write_lock_state = write_lock_idle;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,15,0)
   timer_setup(&pEnumFile->write_lock_timer, (void*)aslenum_write_exclusive_timeout, 0);
#else
   setup_timer(&pEnumFile->write_lock_timer, aslenum_write_exclusive_timeout, (unsigned long)pEnumFile);
#endif

   spin_lock(&baseboard->open_lock);
   if ((file->f_flags & O_EXCL) && (baseboard->open_count > 0)){
      error = -EBUSY;
   }
   else{
      baseboard->open_count++;
      error = 0;
   }
   spin_unlock(&baseboard->open_lock);

   if (error == 0) file->private_data = pEnumFile;

bail:
   if (error){
      dev_err(baseboard->device, "%s: returning %d\n", __FUNCTION__, error);
      if (pEnumFile) kfree(pEnumFile);
   }

   if (g_debug && baseboard) dev_printk(KERN_INFO, baseboard->device, "Done aslenum_open %p %p\n", baseboard, file);

   return error;
}

int aslenum_do_write_exclusive_release(struct aslenum_device *baseboard, struct file * file)
{
   int error = -EIO;
   struct aslenum_file * pEnumFile = file ? file->private_data : NULL;
   u32 signal_exclusive_lock_waiters = 0;

   if (g_debug) printk(KERN_INFO "Entered aslenum_do_write_exclusive_release %p %p\n", baseboard, file);

   if (!baseboard || !pEnumFile) goto nofile;

   error = 0;

   del_timer(&pEnumFile->write_lock_timer);

   spin_lock(&baseboard->open_lock);
   if (baseboard->file_write_exclusive == file){
      baseboard->file_write_exclusive = NULL;
      pEnumFile->write_lock_state = write_lock_idle;
      signal_exclusive_lock_waiters = 1;
   }
   spin_unlock(&baseboard->open_lock);

   if (signal_exclusive_lock_waiters) wake_up(&baseboard->write_exclusive_wait_queue_head);

   if (g_debug) dev_printk(KERN_INFO, baseboard->device, "Done aslenum_do_write_exclusive_release %p %p - %s\n", baseboard, file, signal_exclusive_lock_waiters ? "Released" : "Nothing to do");
   if (error) dev_err(baseboard->device, "%s: returning %d\n", __FUNCTION__, error);

nofile:
   return error;
}

int aslenum_release(struct inode * inode, struct file * file)
{
   struct aslenum_device *baseboard = NULL;
   int error = -EIO;

   baseboard = container_of(inode->i_cdev, struct aslenum_device, cdev);

   if (g_debug) dev_printk(KERN_INFO, baseboard->device, "Entered aslenum_release %p %p\n", baseboard, file);

   spin_lock(&baseboard->open_lock);
   if (baseboard->open_count > 0){
      baseboard->open_count--;
      error = 0;
   }
   else{
      dev_err(baseboard->device, "%s: Device not open\n", __FUNCTION__);
      error = -ENOENT;
   }
   spin_unlock(&baseboard->open_lock);

   aslenum_do_write_exclusive_release(baseboard, file);

   if (error == 0){
      struct aslenum_file *pEnumFile = file->private_data;
      if (pEnumFile){
         del_timer(&pEnumFile->write_lock_timer);
         kfree(pEnumFile);
      }
      file->private_data = NULL;
   }

   if (g_debug) dev_printk(KERN_INFO, baseboard->device, "Done aslenum_release %p %p\n", baseboard, file);
   if (error) dev_err(baseboard->device, "%s: returning %d\n", __FUNCTION__, error);
   return error;
}

void aslenum_write_exclusive_timeout(unsigned long callback_data)
{
   struct aslenum_file * pEnumFile = (struct aslenum_file *)callback_data;
   struct aslenum_device * baseboard = pEnumFile ? pEnumFile->pEnumDevice : NULL;
   u32 signal_timeout = 0;

   if (!pEnumFile || !baseboard) goto bail;

   if (g_debug) dev_printk(KERN_INFO, baseboard->device, "Entered write_exclusive_timeout %p %p\n", baseboard, pEnumFile);

   spin_lock(&baseboard->open_lock);
   switch(pEnumFile->write_lock_state){
      case write_lock_waiting:
         if (g_debug) dev_err(baseboard->device, "state was waiting %p %p\n", baseboard, pEnumFile);
         pEnumFile->write_lock_state = write_lock_timedout;
         signal_timeout = 1;
         break;
      case write_lock_acquired:
         if (g_debug) dev_err(baseboard->device, "state was acquired %p %p\n", baseboard, pEnumFile);
         break;
      default:
         if (g_debug) dev_err(baseboard->device, "state was other %p %p\n", baseboard, pEnumFile);
         pEnumFile->write_lock_state = write_lock_error;
         signal_timeout = 1;
         break;
   }
   spin_unlock(&baseboard->open_lock);

   if (signal_timeout){
      if (g_debug) dev_printk(KERN_INFO, baseboard->device, "waking up waiters %p %p\n", baseboard, pEnumFile);
      wake_up(&baseboard->write_exclusive_wait_queue_head);
   }

   if (g_debug) dev_printk(KERN_INFO, baseboard->device, "Done aslenum_write_exclusive_timeout %p %p\n", baseboard, pEnumFile);

bail:
   return;
}

static int aslenum_exclusive_write_start_wait(struct aslenum_device *baseboard, struct aslenum_file *pEnumFile)
{
   mod_timer(&pEnumFile->write_lock_timer, jiffies + msecs_to_jiffies(pEnumFile->write_lock_timeout_ms));
   return 0;
}

int aslenum_write_exclusive_acquire(struct aslenum_device *baseboard, struct file * file)
{
   int error = -EIO;
   struct aslenum_file * pEnumFile = file ? file->private_data : NULL;
   u32 finished = 0;
   DEFINE_WAIT(wait);

   if (!baseboard || !pEnumFile) goto bail;

   if (g_debug) dev_printk(KERN_INFO, baseboard->device, "Entered write_exclusive_acquire %p %p\n", baseboard, file);

   spin_lock(&baseboard->open_lock);
   pEnumFile->write_lock_state = write_lock_waiting;
   spin_unlock(&baseboard->open_lock);

   while(!finished){
      prepare_to_wait(&baseboard->write_exclusive_wait_queue_head, &wait, TASK_INTERRUPTIBLE);

      if (g_debug) dev_printk(KERN_INFO, baseboard->device, "Looping write_exclusive_acquire %p %p\n", baseboard, file);

      spin_lock(&baseboard->open_lock);
      switch(pEnumFile->write_lock_state){
         case write_lock_waiting:
            if (baseboard->file_write_exclusive == file){
               pEnumFile->write_lock_state = write_lock_error;
               finished = 1;
               error = -EDEADLK;
               if (g_debug) dev_err(baseboard->device, "%s: Deadlock %p %p\n", __FUNCTION__, baseboard, pEnumFile);
            }
            else if (baseboard->file_write_exclusive == NULL){
               baseboard->file_write_exclusive = file;
               pEnumFile->write_lock_state = write_lock_acquired;
               finished = 1;
               error = 0;
            }
            else{
               error = aslenum_exclusive_write_start_wait(baseboard, pEnumFile);
               finished = error ? 1 : 0;
               if (error) dev_err(baseboard->device, "%s: timeout start failed %p %p\n", __FUNCTION__, baseboard, pEnumFile);
            }
            break;
         case write_lock_acquired:
            finished = 1;
            error = 0;
            break;
         case write_lock_timedout:
            finished = 1;
            error = -ETIME;
            if (g_debug) dev_err(baseboard->device, "%s: Timeout %p %p\n", __FUNCTION__, baseboard, pEnumFile);
            break;
         default:
            finished = 1;
            error = -EIO;
            if (g_debug) dev_err(baseboard->device, "%s: Error %p %p\n", __FUNCTION__, baseboard, pEnumFile);
            break;
      }
      spin_unlock(&baseboard->open_lock);
      
      if (finished) break;

      if (signal_pending(current)){
         error = -EAGAIN;
         finished = 1;
         break;
      }
   
      schedule();
   }
   finish_wait(&baseboard->write_exclusive_wait_queue_head, &wait);

   if (error) dev_err(baseboard->device, "%s: returning %d\n", __FUNCTION__, error);
   if (g_debug) dev_printk(KERN_INFO, baseboard->device, "Done write_exclusive_acquire %p %p\n", baseboard, file);
bail:
   return error;
}

int aslenum_write_exclusive_release(struct aslenum_device *baseboard, struct file * file)
{
   int error = 0;

   spin_lock(&baseboard->open_lock);
   if (baseboard->open_count < 1) error = -ENXIO;
   spin_unlock(&baseboard->open_lock);

   if (!error) error = aslenum_do_write_exclusive_release(baseboard, file);

   if (error) dev_err(baseboard->device, "%s: returning %d\n", __FUNCTION__, error);
   return error;
}

int aslenum_check_exclusive(struct aslenum_device *baseboard, struct file * file)
{
   int error = 0;

   spin_lock(&baseboard->open_lock);
   if (baseboard->file_write_exclusive != file) error = -EBUSY;
   spin_unlock(&baseboard->open_lock);
   return error;
}

static const char * aslenum_ioctl_string(unsigned int cmd)
{
   const char * name = "Unknown";

   switch(cmd){
   case MDA_IOCTL_GET_VERSION:
      name = "MDA_IOCTL_GET_VERSION";
      break;
   case MDA_IOCTL_GET_UNIQUE_ID:
      name = "MDA_IOCTL_GET_UNIQUE_ID";
      break;
   case MDAENUM_IOCTL_REGLIST:
      name = "MDA_IOCTL_REGLIST";
      break;
   case MDA_IOCTL_SYSTEM_REGLIST:
      name = "MDA_IOCTL_SYSTEM_REGLIST";
      break;
   case MDAENUM_IOCTL_WRITE_ENABLE:
      name = "MDA_IOCTL_WRITE_ENABLE";
      break;
   case MDAENUM_IOCTL_WRITE_DISABLE:
      name = "MDA_IOCTL_WRITE_DISABLE";
      break;
   case MDAENUM_IOCTL_GET_CHILD_NAME:
      name = "MDA_IOCTL_GET_CHILD_NAME";
      break;
   case MDAENUM_IOCTL_SET_LOCK_TIMEOUT:
      name = "MDA_IOCTL_SET_LOCK_TIMEOUT";
      break;
   case MDAENUM_IOCTL_MANAGE_CHILD_DEVICES:
      name = "MDA_IOCTL_MANAGE_CHILD_DEVICES";
      break;
   default:
      break;
   }
   return name;
}

static long aslenum_ioctl_process(struct file * file, unsigned int cmd, mda_io_buffer *user_io)
{
   long error = -ENXIO;
   struct aslenum_file * pEnumFile = file ? file->private_data : NULL;
   struct aslenum_device * baseboard = pEnumFile ? pEnumFile->pEnumDevice : NULL;

  if (!pEnumFile || !baseboard) goto bail;

   switch(cmd){
   case MDA_IOCTL_GET_VERSION:
      error = mda_ioctl_version(user_io, ASLENUM_VERSION);
      break;
   case MDA_IOCTL_GET_UNIQUE_ID:
      error = aslenum_ioctl_get_unique_id(user_io, baseboard);
      break;
   case MDAENUM_IOCTL_REGLIST:
   case MDA_IOCTL_SYSTEM_REGLIST:
      error = aslenum_ioctl_reglist(baseboard, file, user_io);
      break;

   case MDAENUM_IOCTL_WRITE_ENABLE:
      error = aslenum_write_exclusive_acquire(baseboard, file);
      break;

   case MDAENUM_IOCTL_WRITE_DISABLE:
      error = aslenum_write_exclusive_release(baseboard, file);
      break;

   case MDAENUM_IOCTL_GET_CHILD_NAME:
      error = aslenum_ioctl_get_child_name(baseboard, user_io);
      break;

   case MDAENUM_IOCTL_SET_LOCK_TIMEOUT:
      error = aslenum_ioctl_set_lock_timeout(pEnumFile, user_io);
      break;

   case MDAENUM_IOCTL_MANAGE_CHILD_DEVICES:
      error = 0;
      break;

   default:
      error = -ENOTSUPP;
      break;
   }
   if (error || g_debug) dev_err(baseboard->device, "%s: returning %ld for %s\n", __FUNCTION__, error, aslenum_ioctl_string(cmd));
bail:
   return error;
}

long aslenum_ioctl(struct file * file, unsigned int cmd, unsigned long arg)
{
   long error = 0;
   mda_io_buffer user64_io_buffer;

   memset(&user64_io_buffer, 0, sizeof(user64_io_buffer));
   if (arg != 0){
      error = copy_from_user(&user64_io_buffer, (void*)arg, sizeof(user64_io_buffer));
      if (error) goto bail;
   }
   error = aslenum_ioctl_process(file, cmd, &user64_io_buffer);
bail:
   return error;
}

#if defined __LP64__

long aslenum_compat(struct file * file, unsigned int cmd, unsigned long arg)
{
   long error = 0;
   mda_io_buffer user64_io_buffer;
   struct mda_io32_buffer user32_io_buffer;

   if (arg != 0){
      error = copy_from_user(&user32_io_buffer, (void*)arg, sizeof(user32_io_buffer));
      if (error) goto bail;
      user64_io_buffer.param_buffer.user_address = (void*)(u64)user32_io_buffer.param_buffer.user_address;
      user64_io_buffer.param_buffer.user_size = (size_t)user32_io_buffer.param_buffer.user_size;
      user64_io_buffer.input_buffer.user_address = (void*)(u64)user32_io_buffer.input_buffer.user_address;
      user64_io_buffer.input_buffer.user_size = (size_t)user32_io_buffer.input_buffer.user_size;
      user64_io_buffer.output_buffer.user_address = (void*)(u64)user32_io_buffer.output_buffer.user_address;
      user64_io_buffer.output_buffer.user_size = (size_t)user32_io_buffer.output_buffer.user_size;
   }
   error = aslenum_ioctl_process(file, cmd, &user64_io_buffer);
bail:
   return error;
}
#endif

struct file_operations aslenum_ops = {
   .owner = THIS_MODULE,
   .open  = aslenum_open,
   .release = aslenum_release,
   .unlocked_ioctl = aslenum_ioctl,
#if defined __LP64__
   .compat_ioctl = aslenum_compat,
#endif
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
MODULE_DEVICE_TABLE(file_operations, aslenum_ops);
#else
MODULE_GENERIC_TABLE(file_operations, aslenum_ops);
#endif

