#include "precomp.h"

volatile int aslenum_dma_debug = 0;

static int mda_dma_lock_system(struct mda_device *mda_device, struct mda_dma_buffer *dma_buffer)
{
   int error = -EIO;
   int nPages = 0;
   int page_index = 0;
   int nElements = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0) || MDA_EL8_KERNEL == 1
   if (!access_ok(dma_buffer->user_addr_in, dma_buffer->user_size_in)){
#else
   if (!access_ok(VERIFY_READ, dma_buffer->user_addr_in, dma_buffer->user_size_in)){
#endif
      dev_err(mda_device->parent->parent->parent->device,
              "%s: %s: Buffer not accessible\n",
              mda_device->name,
              __FUNCTION__);
      error = -EFAULT;
      goto bail;
   }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 8, 0)
   mmap_read_lock(current->mm);
#else
   down_read(&current->mm->mmap_sem);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0)
   nPages = get_user_pages(dma_buffer->user_base,
                           dma_buffer->nPages,
                           FOLL_FORCE || dma_buffer->direction == DMA_FROM_DEVICE ? FOLL_WRITE : 0,
                           dma_buffer->page_ptrs,
                           NULL);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 6, 0)
   nPages = get_user_pages(dma_buffer->user_base,
                           dma_buffer->nPages,
                           dma_buffer->direction == DMA_FROM_DEVICE,
                           1, /* Force */
                           dma_buffer->page_ptrs,
                           NULL);
#else
   nPages = get_user_pages(current,
                           current->mm,
                           dma_buffer->user_base,
                           dma_buffer->nPages,
                           dma_buffer->direction == DMA_FROM_DEVICE,
                           1,
                           dma_buffer->page_ptrs,
                           NULL);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 8, 0)
   mmap_read_unlock(current->mm);
#else
   up_read(&current->mm->mmap_sem);
#endif

   if (nPages != dma_buffer->nPages){
      dev_err(mda_device->parent->parent->parent->device,
              "%s: %s get_user_pages failed (%d != %d)\n",
              mda_device->name,
              __FUNCTION__,
              nPages,
              dma_buffer->nPages);
      if (nPages >= 0){
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 8, 0)
         mmap_read_lock(current->mm);
#else
         down_read(&current->mm->mmap_sem);
#endif
         for (page_index = 0; page_index < nPages; page_index++){
           put_page(dma_buffer->page_ptrs[page_index]);
         }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 8, 0)
         mmap_read_unlock(current->mm);
#else
         up_read(&current->mm->mmap_sem);
#endif
         error = -EFAULT;
      }
      goto bail;
   }

   sg_init_table(dma_buffer->scatterlist, nPages);
   for (page_index = 0; page_index < nPages; page_index++){
      sg_set_page(&dma_buffer->scatterlist[page_index],
                  dma_buffer->page_ptrs[page_index],
                  PAGE_SIZE,
                  0);
   }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
   nElements = dma_map_sg(&mda_device->parent->parent->parent->pcidev->dev,
#else
   nElements = pci_map_sg(mda_device->parent->parent->parent->pcidev,
#endif
                          dma_buffer->scatterlist,
                          nPages,
                          dma_buffer->direction == DMA_FROM_DEVICE);
   if (nElements != 0){
      dma_buffer->nSgElements = nElements;
   } else {
      dev_err(mda_device->parent->parent->parent->device,
              "%s: %s ***_map_sg failed to return any elements\n",
              mda_device->name,
              __FUNCTION__);
      error = -EFAULT;
      goto bail;
   }

   error = 0;

bail:
   return error;
}

#if defined MDA_ENABLE_NVIDIA_P2P

#define GPU_BOUND_SHIFT        16
#define GPU_BOUND_SIZE         ((u64)1 << GPU_BOUND_SHIFT)
#define GPU_BOUND_OFFSET       (GPU_BOUND_SIZE-1)
#define GPU_BOUND_MASK         (~GPU_BOUND_OFFSET)

#define PCIEXPRESS_PAGE_SHIFT  12
#define PCIEXPRESS_PAGE_SIZE   ((u64)1 << PCIEXPRESS_PAGE_SHIFT)
#define PCIEXPRESS_PAGE_OFFSET (PCIEXPRESS_PAGE_SIZE - 1)
#define PCIEXPRESS_PAGE_MASK   (~PCIEXPRESS_PAGE_OFFSET)

static void nvidia_p2p_destroy_callback(void *data)
{
   struct mda_dma_buffer *dma_buffer = (struct mda_dma_buffer *)data;
   struct nvidia_p2p_page_table *page_table = NULL;

   down(&dma_buffer->destroy_sem);
   if ((dma_buffer->destroy_callback) && (dma_buffer->destroy_data)){
      dma_buffer->destroy_callback(dma_buffer->destroy_data);
   }
   page_table = xchg(&dma_buffer->nvidia_page_table, NULL);
   if (page_table){
      nvidia_p2p_free_page_table(page_table);
      barrier();
   }
   up(&dma_buffer->destroy_sem);
   return;
}

static void mda_dma_set_sg_page(struct scatterlist *sg_entry, u64 phys_address, u64 length)
{
   sg_entry->offset = (unsigned int)(phys_address & PCIEXPRESS_PAGE_OFFSET);
   sg_entry->length = (unsigned int)length;
   sg_entry->dma_address = (dma_addr_t)phys_address;
#ifdef CONFIG_NEED_SG_DMA_LENGTH
   sg_entry->dma_length = sg_entry->length;
#endif
   return;
}

static int mda_dma_nvidia_pages_to_scatterlist(struct aslenum_device *base_enum, struct mda_dma_buffer *dma_buffer)
{
   struct scatterlist *sg = dma_buffer->scatterlist;                      /* An array of structures */
   u64 nv_count_entries = dma_buffer->nvidia_page_table->entries;
   u64 bytes_remaining = dma_buffer->user_size_in;
   u64 nv_offset = dma_buffer->user_addr_in - dma_buffer->nvidia_va_start;
   const u64 sg_page_size = PAGE_SIZE;
   const u64 nv_page_size = dma_buffer->nvidia_page_size;
   u64 nv_index = 0;
   u64 sg_index = 0;
   u64 phys_base_address = dma_buffer->nvidia_page_table->pages[nv_index]->physical_address;
   u64 sg_bytes_this_page = sg_page_size - (dma_buffer->user_addr_in & (sg_page_size - 1));
   u64 nv_bytes_remaining = nv_page_size - nv_offset;

   if (aslenum_dma_debug){
      dev_err(base_enum->device, "%s: entered with %d nvidia pages of %d bytes to scatterlist %p\n",
              __FUNCTION__,
              (int)nv_count_entries,
              (int)nv_page_size,
              (void*)sg);
      for(nv_index = 0; nv_index < nv_count_entries; nv_index++)
      {
         dev_err(base_enum->device, "%s: In - nv address %d = %p\n", __FUNCTION__, (int)nv_index, (void*)dma_buffer->nvidia_page_table->pages[nv_index]->physical_address);
      }
   }

   memset(sg, 0, dma_buffer->nPages * sizeof(*sg));
   sg_init_table(sg, dma_buffer->nPages);

   if (aslenum_dma_debug){
      for(nv_index = 0; nv_index < nv_count_entries; nv_index++)
      {
         dev_err(base_enum->device, "%s: In - nv address %d = %p\n", __FUNCTION__, (int)nv_index, (void*)dma_buffer->nvidia_page_table->pages[nv_index]->physical_address);
      }
   }
   nv_index = 0;

   while(bytes_remaining > 0){
      u64 phys_address;

      if (aslenum_dma_debug){
         dev_err(base_enum->device, "%s: remain = %lu, sg this = %lu, nv remain = %lu, sg = %lu, nv = %lu",
                 __FUNCTION__,
                 (long unsigned int)bytes_remaining,
                 (long unsigned int)sg_bytes_this_page,
                 (long unsigned int)nv_bytes_remaining,
                 (long unsigned int)sg_index,
                 (long unsigned int)nv_index); 
      }

      if (nv_bytes_remaining == 0){
         if (++nv_index >= nv_count_entries) break;
         nv_bytes_remaining = nv_page_size;
         phys_base_address = dma_buffer->nvidia_page_table->pages[nv_index]->physical_address;
         nv_offset = 0;
         if (aslenum_dma_debug){
            dev_err(base_enum->device, "%s: New nv - phys[%d] = %p\n",
                    __FUNCTION__,
                    (int)nv_index,
                    (void*)phys_base_address);
         }
      }

      if (sg_bytes_this_page > bytes_remaining) sg_bytes_this_page = bytes_remaining;
      if (sg_bytes_this_page > sg_page_size) sg_bytes_this_page = sg_page_size;
      if (sg_bytes_this_page > nv_bytes_remaining) sg_bytes_this_page = nv_bytes_remaining;

      phys_address = phys_base_address + nv_offset;
      mda_dma_set_sg_page(&sg[sg_index], phys_address, sg_bytes_this_page);
      if (aslenum_dma_debug){
         dev_err(base_enum->device, "%s: set page %lu (%p) to %p %lu\n",
                 __FUNCTION__,
                 (long unsigned int)sg_index,
                 &sg[sg_index],
                 (void*)phys_address,
                 (long unsigned int)sg_bytes_this_page);
      }

      bytes_remaining -= sg_bytes_this_page;
      nv_bytes_remaining -= sg_bytes_this_page;
      nv_offset += sg_bytes_this_page;

      if (bytes_remaining){
         sg_index++;
      }
      sg_bytes_this_page = bytes_remaining;
   }
   sg_mark_end(&sg[sg_index]);
   dma_buffer->nSgElements = sg_index + 1;
   if (aslenum_dma_debug){
      dev_err(base_enum->device, "%s: bytes remaining = %lu, entries written = %lu\n", __FUNCTION__, (long unsigned int)bytes_remaining, (long unsigned int)dma_buffer->nSgElements);
      for(nv_index = 0; nv_index < nv_count_entries; nv_index++)
      {
         dev_err(base_enum->device, "%s: Out - nv address %d = %p\n", __FUNCTION__, (int)nv_index, (void*)dma_buffer->nvidia_page_table->pages[nv_index]->physical_address);
      }
   }
   return 0;
}

static int mda_dma_lock_nvidia_p2p(struct mda_device *mda_device, struct mda_dma_buffer *dma_buffer)
{
   int status = -EINVAL;
   struct aslenum_device *base_enum = mda_device->parent->parent->parent;

   sema_init(&dma_buffer->destroy_sem, 0);

   if (!dma_buffer->user_addr_in) goto bail;
   if (!dma_buffer->user_size_in) goto bail;

   dma_buffer->nvidia_va_start = dma_buffer->user_addr_in & GPU_BOUND_MASK;
   dma_buffer->nvidia_mapped_size =  (dma_buffer->user_addr_in - dma_buffer->nvidia_va_start) +  dma_buffer->user_size_in;
   dma_buffer->nvidia_page_table = NULL;

   if (!dma_buffer->nvidia_va_start) goto bail;
   if (!dma_buffer->nvidia_mapped_size) goto bail;

   if (aslenum_dma_debug){
      dev_err(base_enum->device,
              "%s: %s token %lu, space %d, va_start %p, size %lu\n",
              mda_device->name,
              __FUNCTION__,
              (long unsigned int)dma_buffer->nvidia_p2p_token,
              dma_buffer->nvidia_va_space,
              (void*)dma_buffer->nvidia_va_start,
              (long unsigned int)dma_buffer->nvidia_mapped_size);
   }

   status = nvidia_p2p_get_pages(dma_buffer->nvidia_p2p_token,
                                 dma_buffer->nvidia_va_space,
                                 dma_buffer->nvidia_va_start,
                                 dma_buffer->nvidia_mapped_size,
                                 &dma_buffer->nvidia_page_table,
                                 nvidia_p2p_destroy_callback,
                                 dma_buffer);
   if (status){
      dev_err(base_enum->device,
              "%s: %s nvidia_p2p_get_pages failed with %d\n",
              mda_device->name,
              __FUNCTION__,
              status);
      goto bail; 
   }

   if (!dma_buffer->nvidia_page_table){
      dev_err(base_enum->device,
              "%s: %s nvidia_p2p_get_pages returned NULL page table\n",
              mda_device->name,
              __FUNCTION__);
      status = -EFAULT;
      goto bail; 
   }
   switch(dma_buffer->nvidia_page_table->page_size){
   case NVIDIA_P2P_PAGE_SIZE_4KB:
      dma_buffer->nvidia_page_size = 4 * 1024;
      break;
   case NVIDIA_P2P_PAGE_SIZE_64KB:
      dma_buffer->nvidia_page_size = 64 * 1024;
      break;
   case NVIDIA_P2P_PAGE_SIZE_128KB:
      dma_buffer->nvidia_page_size = 128 * 1024;
      break;
   default:
      dev_err(base_enum->device,
              "%s: %s nvidia_p2p_get_pages returned invalid page size\n",
              mda_device->name,
              __FUNCTION__);
      status = -EFAULT;
      goto bail;
      break;
   }
   
   status = mda_dma_nvidia_pages_to_scatterlist(base_enum, dma_buffer);
bail:
   up(&dma_buffer->destroy_sem);
   return status;;
}
#endif

int mda_dma_lock(struct mda_device *mda_device, struct mda_dma_buffer *dma_buffer)
{
   int status;

   switch(dma_buffer->memory_type){
      case system_memory:
         status = mda_dma_lock_system(mda_device, dma_buffer);
         break;
#if defined MDA_ENABLE_NVIDIA_P2P
      case nvidia_p2p_memory:
         status = mda_dma_lock_nvidia_p2p(mda_device, dma_buffer);
         break;
#endif
      default:
         status = -ENOTSUPP;
         break;
   }
   return status;
}

static void mda_dma_unlock_system(struct mda_device *mda_device, struct mda_dma_buffer *dma_buffer)
{
   int page_index = 0;
   int mmvalid = 0;

   if (current){
      if (current->mm){
         mmvalid = 1;
      }
   }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
   dma_unmap_sg(&mda_device->parent->parent->parent->pcidev->dev,
#else
   pci_unmap_sg(mda_device->parent->parent->parent->pcidev,
#endif
                dma_buffer->scatterlist,
                dma_buffer->nPages,
                dma_buffer->direction == DMA_FROM_DEVICE);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 8, 0)
   if (mmvalid) mmap_read_lock(current->mm);
#else
   if (mmvalid) down_read(&current->mm->mmap_sem);
#endif
   for (page_index = 0; page_index < dma_buffer->nPages; page_index++){
      put_page(dma_buffer->page_ptrs[page_index]);
   }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 8, 0)
   if (mmvalid) mmap_read_unlock(current->mm);
#else
   if (mmvalid) up_read(&current->mm->mmap_sem);
#endif

   return;
}

#if defined MDA_ENABLE_NVIDIA_P2P
static void mda_dma_unlock_nvidia_p2p(struct mda_device *mda_device, struct mda_dma_buffer *dma_buffer)
{
   struct nvidia_p2p_page_table * page_table = NULL;

   page_table = xchg(&dma_buffer->nvidia_page_table, NULL);
   if (page_table){
      nvidia_p2p_put_pages(dma_buffer->nvidia_p2p_token,
                           dma_buffer->nvidia_va_space,
                           dma_buffer->nvidia_va_start,
                           page_table);
      barrier();
   }
   return;
}
#endif

void mda_dma_unlock(struct mda_device *mda_device, struct mda_dma_buffer *dma_buffer)
{
   switch(dma_buffer->memory_type){
      case system_memory:
         mda_dma_unlock_system(mda_device, dma_buffer);
         break;
#if defined MDA_ENABLE_NVIDIA_P2P
      case nvidia_p2p_memory:
         mda_dma_unlock_nvidia_p2p(mda_device, dma_buffer);
         break;
#endif
      default:
         break;
   }
   return;
}

EXPORT_SYMBOL(mda_dma_lock);
EXPORT_SYMBOL(mda_dma_unlock);

