#include "precomp.h"

extern struct file_operations aslenum_ops;

static struct aslenum_driver_struct {
   struct class * class;
   dev_t          driver_dev;

   spinlock_t     lock;
   dev_t          next_minor;

} aslenum_driver;

static dev_t aslenum_driver_get_next_minor(struct aslenum_device *firebird)
{
   dev_t minor = 0;

   spin_lock(&aslenum_driver.lock);
   minor = aslenum_driver.next_minor++;
   spin_unlock(&aslenum_driver.lock);

   return minor;
}

static int MDA_DEVINIT aslenum_config_space_cache(struct aslenum_device *firebird)
{
   int error = -EIO;
   uint32_t index = 0;
   uint32_t dword_total = sizeof(firebird->config_space_cache) / sizeof(uint32_t);

   while(index < dword_total){
      uint32_t config_offset = index * sizeof(uint32_t);
      uint32_t *cache_address = (uint32_t*)&(firebird->config_space_cache[config_offset]);
      error = pci_read_config_dword( firebird->pcidev, config_offset, cache_address );
      if (error){
         dev_err(firebird->device, "%s: pci_read_config_dword failed with %d\n", __FUNCTION__, error);
         goto bail;
      }
      index++;
   }
bail:
   return error;
}

int aslenum_device_probe( struct pci_dev* pcidev, const struct pci_device_id * template)
{
   int error = -ENXIO;
   struct aslenum_device *firebird = NULL;
   char *device_name = "Unknown";
   enum baseboard_type baseboard_type = board_type_unknown;
   int io_bar_present = 0;

   switch( pcidev->device  ){
   case FBD_V6_2PE8_DEVICE_ID:
   case FBD_A7_2PE4_DEVICE_ID:
   case FBD_KU_3PE4_DEVICE_ID:
   case FBD_KU_3PE8_DEVICE_ID:
      baseboard_type = board_type_FireBird;
      device_name = ASLENUM_STRING_FIREBIRD;
      io_bar_present = 0;
      break;
   case PHX_S3_1PEX_DEVICE_ID:
      baseboard_type = board_type_Phoenix;
      device_name = ASLENUM_STRING_PHOENIX;
      io_bar_present = 1;
      break;
   default:
      goto err_unrecognised;
      break;
   }

   firebird = kzalloc(sizeof(*firebird), GFP_KERNEL);
   if (firebird == NULL){
      dev_err(&pcidev->dev, "%s: kzalloc failed\n", __FUNCTION__);
      error = -ENOMEM;
      goto err_kzalloc;
   }
   firebird->baseboard_type = baseboard_type;
   firebird->pcidev = pcidev;
   firebird->major = MAJOR( aslenum_driver.driver_dev );
   firebird->minor = aslenum_driver_get_next_minor(firebird);
   firebird->owner = THIS_MODULE;
   INIT_LIST_HEAD(&firebird->mda_root_node_list);
   spin_lock_init(&firebird->open_lock);
   spin_lock_init(&firebird->mda_root_node_lock);
   init_waitqueue_head(&firebird->write_exclusive_wait_queue_head);

   error = pci_enable_device( pcidev );
   if (error){
      dev_err(&pcidev->dev, "%s: pci_enable_device failed with %d\n", __FUNCTION__, error);
      goto err_enable;
   }
   pci_set_master( pcidev );

   error = pci_request_regions( pcidev, ASLENUM_DEVICE_NAME );
   if (error){
      dev_err(&pcidev->dev, "%s: pci_request_regions failed with %d\n", __FUNCTION__, error);
      goto err_regions;
   }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
   error = dma_set_mask(&pcidev->dev, DMA_BIT_MASK(64));
#else
   error = pci_set_dma_mask(pcidev, DMA_BIT_MASK(64));
#endif
   if (error){
      dev_err(&pcidev->dev, "%s: ***_set_dma_mask failed with %d\n", __FUNCTION__, error);
      goto err_dma_mask;
   }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 18, 0)
   error = dma_set_coherent_mask(&pcidev->dev, DMA_BIT_MASK(64));
#else
   error = pci_set_consistent_dma_mask(pcidev, DMA_BIT_MASK(64));
#endif
   if (error){
      dev_err(&pcidev->dev, "%s: ***_set_consistent_dma_mask failed with %d\n", __FUNCTION__, error);
      goto err_dma_mask;
   }

   firebird->board_address.kvaddr = pci_iomap(pcidev, 0, 0);
   if (firebird->board_address.kvaddr == NULL){
      dev_err(&pcidev->dev, "%s: pci_iomap memory failed\n", __FUNCTION__);
      error = -ENOMEM;
      goto err_iomap_mem;
   }
   firebird->board_address.kvsize = pci_resource_len(pcidev, 0);
   firebird->device_address.kvaddr = firebird->board_address.kvaddr;
   firebird->device_address.kvsize = firebird->board_address.kvsize;

   if (io_bar_present){
      firebird->ioport_address.kvaddr = pci_iomap(pcidev, 1, 0);
      if (firebird->ioport_address.kvaddr == NULL){
         dev_err(&pcidev->dev, "%s: pci_iomap I/O failed\n", __FUNCTION__);
         error = -ENOMEM;
         goto err_iomap_ioport;
      }
      firebird->ioport_address.kvsize = pci_resource_len(pcidev, 1);
   }

   pci_set_drvdata( pcidev, firebird );

   error = request_irq( pcidev->irq, aslenum_isr, IRQF_SHARED, ASLENUM_DEVICE_NAME, firebird);
   if (error){
      dev_err(&pcidev->dev, "%s: request_irq failed with %d\n", __FUNCTION__, error);
      goto err_irq;
   }

   error = aslenum_config_space_cache(firebird);
   if (error){
      dev_err(&pcidev->dev, "%s: failed to cache config space %d\n", __FUNCTION__, error);
      goto err_config_space;
   }

   cdev_init(&firebird->cdev, &aslenum_ops);
   firebird->cdev.owner = firebird->owner;

   error = cdev_add(&firebird->cdev, MKDEV(firebird->major, firebird->minor), 1);
   if (error){
      dev_err(&pcidev->dev, "%s: cdev_add failed with %d\n", __FUNCTION__, error);
      goto err_cdev;
   }

   firebird->device = device_create(aslenum_driver.class,
                                    &pcidev->dev,
                                    MKDEV(firebird->major, firebird->minor),
                                    NULL,
                                    "%s%d", ASLENUM_DEVICE_NAME, firebird->minor);
   if (firebird->device == NULL){
      dev_err(&pcidev->dev, "%s: pci_iomap failed\n", __FUNCTION__);
      error = -ENODEV;
      goto err_device;
   }

   dev_printk(KERN_INFO,
              firebird->device,
              "%s 0x%04X/0x%04X 0x%04X/0x%04X Rev %u\n",
              device_name,
              pcidev->vendor,
              pcidev->device,
              pcidev->subsystem_vendor,
              pcidev->subsystem_device,
              pcidev->revision);
   dev_printk(KERN_INFO,
              firebird->device,
              "registered at %u/%u\n",
              firebird->major,
              firebird->minor);
   dev_printk(KERN_INFO, 
              firebird->device,
              "BAR0 = %p (%08X), irqpin = %u, irq = %u\n",
              firebird->board_address.kvaddr,
              (u32)pci_resource_start( pcidev, 0 ),
              pcidev->pin,
              pcidev->irq);

   error = mda_create_nodes(firebird);
   if (error){
      dev_err(firebird->device, "%s: mda_create_nodes failed with %d\n", __FUNCTION__, error);
      goto err_mda_root_node;
   }

   if (!io_bar_present){
      dev_printk(KERN_INFO,
                 firebird->device,
                 "%s: Firmware version 0x%08X\n",
                 __FUNCTION__,
                __raw_readl(firebird->device_address.kvaddr + 4));
   }

   error = 0;
   goto finished;

err_mda_root_node:
   device_destroy(aslenum_driver.class, MKDEV(firebird->major, firebird->minor));
err_device:
   cdev_del(&firebird->cdev);
err_cdev:
err_config_space:
   free_irq(pcidev->irq, firebird);
err_irq:
   if (io_bar_present) pci_iounmap( pcidev, firebird->ioport_address.kvaddr );
err_iomap_ioport:
   pci_iounmap( pcidev, firebird->board_address.kvaddr );
err_iomap_mem:
   pci_release_regions( pcidev );
err_dma_mask:
err_regions:
   pci_disable_device( pcidev );
err_enable:
   kfree(firebird);
err_kzalloc:
err_unrecognised:
finished:
   return error;
}

void aslenum_device_remove(struct pci_dev * pcidev)
{
   struct aslenum_device *firebird = pci_get_drvdata(pcidev);

   dev_printk(KERN_INFO, 
              firebird->device,
              "unloading 0x%04X/0x%04X 0x%04X/0x%04X Rev %u\n",
              pcidev->vendor,
              pcidev->device,
              pcidev->subsystem_vendor,
              pcidev->subsystem_device,
              pcidev->revision);
   mda_destroy_nodes(firebird);
   device_destroy(aslenum_driver.class, MKDEV(firebird->major, firebird->minor));
   cdev_del(&firebird->cdev);
   free_irq( pcidev->irq, firebird );
   if (firebird->ioport_address.kvaddr) pci_iounmap(pcidev, firebird->ioport_address.kvaddr);
   pci_iounmap(pcidev, firebird->board_address.kvaddr);
   pci_release_regions(pcidev);
   pci_disable_device(pcidev);
   kfree(firebird);
   return;
}

int aslenum_device_suspend(struct pci_dev* pcidev, pm_message_t state)
{
   struct aslenum_device *firebird = pci_get_drvdata(pcidev);

   dev_printk(KERN_INFO, firebird->device, "%s\n", __FUNCTION__);
   return 0;
}

int aslenum_device_suspend_late(struct pci_dev * pcidev, pm_message_t state)
{
   struct aslenum_device *firebird = pci_get_drvdata(pcidev);

   dev_printk(KERN_INFO, firebird->device, "%s\n", __FUNCTION__);
   return 0;
}

int aslenum_device_resume(struct pci_dev * pcidev)
{
   struct aslenum_device *firebird = pci_get_drvdata(pcidev);

   dev_printk(KERN_INFO, firebird->device, "%s\n", __FUNCTION__);
   return 0;
}

int aslenum_device_resume_early(struct pci_dev * pcidev)
{
   struct aslenum_device *firebird = pci_get_drvdata(pcidev);

   dev_printk(KERN_INFO, firebird->device, "%s\n", __FUNCTION__);
   return 0;
}

void aslenum_device_shutdown(struct pci_dev * pcidev)
{
   struct aslenum_device *firebird = pci_get_drvdata(pcidev);

   dev_printk(KERN_INFO, firebird->device, "%s\n", __FUNCTION__);
   return;
}

struct pci_device_id aslenum_pci_table[] MDA_DEVINITDATA =
{
   {
       FBD_XILINX_VENDOR_ID,                   /* Xilinx */
       FBD_V6_2PE8_DEVICE_ID,                  /* Spartan6 */
       FBD_SUBVENDOR_ID,                       /* ASL */
       PCI_ANY_ID,                             /* All FireBird */
       board_type_FireBird                     /* Flag for device probe */
   },
   {
       FBD_XILINX_VENDOR_ID,                   /* Xilinx */
       FBD_A7_2PE4_DEVICE_ID,                  /* Artix7 */
       FBD_SUBVENDOR_ID,                       /* ASL */
       PCI_ANY_ID,                             /* All FireBird */
       board_type_FireBird                     /* Flag for device probe */
   },
   {
       FBD_XILINX_VENDOR_ID,                   /* Xilinx */
       FBD_KU_3PE4_DEVICE_ID,                  /* Kintex UltraScale */
       FBD_SUBVENDOR_ID,                       /* ASL */
       PCI_ANY_ID,                             /* All FireBird */
       board_type_FireBird                     /* Flag for device probe */
   },
   {
       FBD_XILINX_VENDOR_ID,                   /* Xilinx */
       FBD_KU_3PE8_DEVICE_ID,                  /* Kintex UltraScale */
       FBD_SUBVENDOR_ID,                       /* ASL */
       PCI_ANY_ID,                             /* All FireBird */
       board_type_FireBird                     /* Flag for device probe */
   },
   {
       FBD_XILINX_VENDOR_ID,                   /* Xilinx */
       PHX_S3_1PEX_DEVICE_ID,                  /* Spartan3 */
       FBD_SUBVENDOR_ID,                       /* ASL */
       PCI_ANY_ID,                             /* All Phoenix */
       board_type_Phoenix                      /* Flag for device probe */
   },
   { 0 }                                       /* Terminate with empty struct */
};

struct pci_driver aslenum_pci_driver = {
   .name         = ASLENUM_DEVICE_NAME,
   .probe        = aslenum_device_probe,
   .remove       = aslenum_device_remove,
   .suspend      = aslenum_device_suspend,
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,5,0) && MDA_EL8_KERNEL != 1
   .suspend_late = aslenum_device_suspend_late,
#endif
   .resume       = aslenum_device_resume,
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,5,0) && MDA_EL8_KERNEL != 1
   .resume_early = aslenum_device_resume_early,
#endif
   .shutdown     = aslenum_device_shutdown,
   .id_table     = aslenum_pci_table,
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
static CLASS_ATTR_STRING(version, S_IRUGO, ASLENUM_VERSION_STRING);
#endif

static int __init aslenum_driver_init(void)
{
   int error = 0;

   printk(KERN_INFO "%s: Built on %s at %s\n", __FUNCTION__, __DATE__, __TIME__);

   memset(&aslenum_driver, 0, sizeof(aslenum_driver));
   spin_lock_init(&aslenum_driver.lock);

   aslenum_driver.class = class_create(THIS_MODULE, ASLENUM_DEVICE_NAME);
   if (!aslenum_driver.class){
      error = -ENOMEM;
      printk(KERN_ERR "%s class_create returned %d\n", __FUNCTION__, error);
      goto err_class;
   }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
   error = class_create_file(aslenum_driver.class, &class_attr_version.attr );
   if (error)
   {
      printk(KERN_ERR "%s class_create_file returned %d\n", __FUNCTION__, error);
      goto err_attr;
   }
#endif
   error = aslenum_bus_init();
   if (error){
      printk(KERN_ERR "%s: aslenum_bus_init failed with %d\n", __FUNCTION__, error);
      goto err_bus;
   }

   error = alloc_chrdev_region(&aslenum_driver.driver_dev, 0, ASLENUM_MAX_DEVICES, ASLENUM_DEVICE_NAME);
   if (error){
      printk(KERN_ERR "%s alloc_chrdev_region returned %d\n", __FUNCTION__, error);
      goto err_chr;
   }
   printk(KERN_INFO "%s: major number = %u\n", __FUNCTION__, MAJOR(aslenum_driver.driver_dev));

   error = pci_register_driver(&aslenum_pci_driver);
   if (error){
      printk(KERN_ERR "%s pci_register_driver returned %d\n", __FUNCTION__, error);
      goto err_pci;
   }

   goto finished;

err_pci:
   unregister_chrdev_region( aslenum_driver.driver_dev, ASLENUM_MAX_DEVICES );
err_chr:
   aslenum_bus_remove();
err_bus:
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
   class_remove_file(aslenum_driver.class, &class_attr_version.attr );
#endif
err_attr:
   class_destroy(aslenum_driver.class);
err_class:
finished:

   return error;
}

void aslenum_driver_cleanup(void)
{
   printk(KERN_INFO "%s\n", __FUNCTION__);

   pci_unregister_driver(&aslenum_pci_driver);
   unregister_chrdev_region(aslenum_driver.driver_dev, ASLENUM_MAX_DEVICES );
   aslenum_bus_remove();
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
   class_remove_file(aslenum_driver.class, &class_attr_version.attr);
#endif
   class_destroy(aslenum_driver.class);
   return;
}

module_init(aslenum_driver_init);
module_exit(aslenum_driver_cleanup);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
MODULE_DEVICE_TABLE(pci, aslenum_pci_table);
#else
MODULE_GENERIC_TABLE(pci, aslenum_pci_table);
#endif
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Chris Read <techsupport@activesilicon.com>");


