#include "precomp.h"

static struct aslser_driver_struct {
   int           port_index;
   spinlock_t    lock;
} aslser_driver;

static struct uart_driver aslser_uart_driver = {
   .owner = THIS_MODULE,
   .driver_name = ASLSER_DRIVER_NAME,
   .dev_name = ASLSER_DEVICE_NAME,
   .nr = ASLSER_MAX_DEVICES,
   .major = 0, /* Filled in by init routine */
   .minor = 0, /* Filled in by init routine */
   .cons = NULL,
   .tty_driver = NULL,
   .state = NULL
};

static dev_t aslser_driver_get_next_port_index(struct aslser_device *ser_device)
{
   dev_t minor = 0;

   spin_lock(&aslser_driver.lock);
   minor = aslser_driver.port_index++;
   spin_unlock(&aslser_driver.lock);

   return minor;
}

static int aslser_device_probe(struct mda_device *mdadev, const struct mda_device_id *template)
{
   int error = -ENXIO;
   struct aslser_device *ser_device = NULL;

   ser_device = kzalloc(sizeof(*ser_device), GFP_KERNEL);
   if (ser_device == NULL){
      printk(KERN_ERR "%s: kzalloc (device) failed\n", __FUNCTION__);
      error = -ENOMEM;
      goto err_kzalloc_device;
   }

   ser_device->mdadev = mdadev;
   ser_device->owner = THIS_MODULE;
   ser_device->device = get_device(&mdadev->device);

   ser_device->device_id.driver_matching_name = template->driver_matching_name;
   ser_device->device_id.device_name_type = template->device_name_type;

   switch(template->device_name_type){
   case ASLSER_DEVICE_TYPE_CAMERALINK:
      ser_device->uart_port.ops = &fbclser_ops;
      ser_device->uart_port.fifosize = FBD_SERIAL_TX_FIFO_MAX_DEPTH;
      ser_device->uart_port.uartclk = FBD_SERIAL_BAUD_CLOCK;
      ser_device->uart_port.type = PORT_UNKNOWN;
      ser_device->uart_port.flags = UPF_BOOT_AUTOCONF | UPF_SKIP_TEST | UPF_SHARE_IRQ;
      break;
   default:
      printk(KERN_ERR "%s: Unrecognised device %d\n", __FUNCTION__, (int)template->device_name_type);
      error = -EIO;
      goto err_unknown_device;
   }

   ser_device->uart_port.line = aslser_driver_get_next_port_index(ser_device);
   ser_device->uart_port.dev = ser_device->device;
   ser_device->uart_port.mapbase = pci_resource_start(mdadev->parent->parent->parent->pcidev, 0);
   ser_device->uart_port.mapbase += mdadev->params.memory_offset;
   ser_device->uart_port.membase = mdadev->device_address.kvaddr;
   ser_device->uart_port.iotype = UPIO_MEM;

   mda_set_drvdata(mdadev, ser_device);

   error = uart_add_one_port(&aslser_uart_driver, &ser_device->uart_port);
   if (error){
      printk(KERN_ERR "%s: uart_add_one_port returned %d\n", __FUNCTION__, error);
      goto err_uart_add;
   }

   sprintf(mdadev->name, "%s%d", aslser_uart_driver.dev_name, ser_device->uart_port.line);

   dev_printk(KERN_INFO,
              &ser_device->mdadev->device,
              "%s: Registered %s\n",
              __FUNCTION__,
              mdadev->name);

   error = 0;
   goto finished;

err_uart_add:
err_unknown_device:
   put_device(ser_device->device);
   kfree(ser_device);
err_kzalloc_device:
finished:
   return error;
}

static void aslser_device_remove(struct mda_device *mdadev)
{
   struct aslser_device *ser_device = mda_get_drvdata(mdadev);

   dev_printk(KERN_INFO, ser_device->device, "%s: unloading\n", __FUNCTION__);
   uart_remove_one_port(&aslser_uart_driver, &ser_device->uart_port);
   put_device(ser_device->device);
   kfree(ser_device);
   return;
}

static int aslser_device_suspend(struct mda_device *mdadev, pm_message_t state)
{
   struct aslser_device *ser_device = mda_get_drvdata(mdadev);

   if (ser_device) dev_printk(KERN_INFO, ser_device->device, "%s\n", __FUNCTION__);
   else printk(KERN_INFO "%s\n", __FUNCTION__);
   return 0;
}

static int aslser_device_suspend_late(struct mda_device *mdadev, pm_message_t state)
{
   struct aslser_device *ser_device = mda_get_drvdata(mdadev);

   if (ser_device) dev_printk(KERN_INFO, ser_device->device, "%s\n", __FUNCTION__);
   else printk(KERN_INFO "%s\n", __FUNCTION__);
   return 0;
}

static int aslser_device_resume(struct mda_device *mdadev)
{
   struct aslser_device *ser_device = mda_get_drvdata(mdadev);

   if (ser_device) dev_printk(KERN_INFO, ser_device->device, "%s\n", __FUNCTION__);
   else printk(KERN_INFO "%s\n", __FUNCTION__);
   return 0;
}

static int aslser_device_resume_early(struct mda_device *mdadev)
{
   struct aslser_device *ser_device = mda_get_drvdata(mdadev);

   if (ser_device) dev_printk(KERN_INFO, ser_device->device, "%s\n", __FUNCTION__);
   else printk(KERN_INFO "%s\n", __FUNCTION__);
   return 0;
}

static void aslser_device_shutdown(struct mda_device *mdadev)
{
   struct aslser_device *ser_device = mda_get_drvdata(mdadev);

   if (ser_device) dev_printk(KERN_INFO, ser_device->device, "%s\n", __FUNCTION__);
   else printk(KERN_INFO "%s\n", __FUNCTION__);
   return;
}

struct mda_device_id aslser_mda_table[] =
{
   {
      "aslser",
      ASLSER_DEVICE_TYPE_CAMERALINK
   },
   { 0 }                                       /* Terminate with empty struct */
};

struct mda_driver aslser_mda_driver = {
   .name         = ASLSER_DEVICE_NAME,
   .probe        = aslser_device_probe,
   .remove       = aslser_device_remove,
   .suspend      = aslser_device_suspend,
   .suspend_late = aslser_device_suspend_late,
   .resume       = aslser_device_resume,
   .resume_early = aslser_device_resume_early,
   .shutdown     = aslser_device_shutdown,
   .id_table     = aslser_mda_table,
};

static int __init aslser_driver_init(void)
{
   int error = 0;

   printk(KERN_INFO "%s: Built on %s at %s\n", __FUNCTION__, __DATE__, __TIME__);

   memset(&aslser_driver, 0, sizeof(aslser_driver));

   spin_lock_init(&aslser_driver.lock);

   error = uart_register_driver(&aslser_uart_driver);
   if (error){
      printk(KERN_ERR "%s uart_register_driver returned %d\n", __FUNCTION__, error);
      goto err_uart_register;
   }

   error = mda_register_driver(&aslser_mda_driver);
   if (error){
      printk(KERN_ERR "%s mda_register_driver returned %d\n", __FUNCTION__, error);
      goto err_mda;
   }

   goto finished;

err_mda:
   uart_unregister_driver(&aslser_uart_driver);
err_uart_register:
finished:
   return error;
}

static void __exit aslser_driver_cleanup(void)
{
   printk(KERN_INFO "%s\n", __FUNCTION__);

   mda_unregister_driver(&aslser_mda_driver);
   uart_unregister_driver(&aslser_uart_driver);

   return;
}

module_init(aslser_driver_init);
module_exit(aslser_driver_cleanup);
MODULE_DEVICE_TABLE(mda, aslser_mda_table);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Chris Read <techsupport@activesilicon.com>");
