#include "precomp.h"

int aslser_fbcl_debug = 0;

static void aslser_fbcl_printk(struct aslser_device *ser_device, const char *severity, const char *format, ...)
{
   va_list va;
   char *output_string = NULL;
   size_t size_string = 1024;

   if (!aslser_fbcl_debug) goto bail;
   output_string = kzalloc(size_string, GFP_KERNEL);
   if (!output_string) goto bail;

   va_start(va, format);
   vsnprintf(output_string, size_string, format, va);
   va_end(va);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
   dev_printk(severity, &ser_device->mdadev->device,
#else
   dev_printk(KERN_INFO, &ser_device->mdadev->device,
#endif
                        "%s%d: %s",
                        ASLSER_DEVICE_NAME,
                        ser_device->uart_port.line,
                        output_string);
bail:
   if (output_string) kfree(output_string);
   return;
}

static void aslser_fbcl_debug_entry(struct aslser_device *ser_device, const char *function)
{
   if (aslser_fbcl_debug) aslser_fbcl_printk(ser_device, KERN_INFO, "%s: entered\n", function);
   return;
}

/* called unlocked from this file */
static void aslser_fbcl_reset_port(struct aslser_device *ser_device)
{
   unsigned long irq_state = 0;

   spin_lock_irqsave(&ser_device->uart_port.lock, irq_state);
   ser_device->shadow_interrupt_control = 0;
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_INT_CTRL, ser_device->shadow_interrupt_control);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_INT_CTRL, ser_device->shadow_serial_control | FBD_SERIAL_BITS_CTRL_GLOBAL_RESET);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_INT_CTRL, ser_device->shadow_serial_control);
   spin_unlock_irqrestore(&ser_device->uart_port.lock, irq_state);

   return;
}

static void fbclser_tx_off(struct aslser_device *ser_device)
{
   ser_device->shadow_interrupt_control &= ~FBD_SERIAL_BITS_INT_CTRL_TXEMPTY_ENABLE;
   ser_device->shadow_interrupt_control &= ~FBD_SERIAL_BITS_INT_CTRL_TXHALF_ENABLE;
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_INT_CTRL, ser_device->shadow_interrupt_control);
   return;
}

static void fbclser_tx_on(struct aslser_device *ser_device)
{
   ser_device->shadow_interrupt_control |= FBD_SERIAL_BITS_INT_CTRL_TXEMPTY_ENABLE;
   ser_device->shadow_interrupt_control |= FBD_SERIAL_BITS_INT_CTRL_GLOBAL_ENABLE;
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_INT_CTRL, ser_device->shadow_interrupt_control);
   return;
}

/* Intermediate inline functions to reduce conditional compilation */
static inline int aslser_tty_insert_flip_char(struct aslser_device *ser_device, unsigned char ch, char flag)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,9,0)
   return tty_insert_flip_char(&ser_device->uart_port.state->port, ch, flag);
#else
   return tty_insert_flip_char(ser_device->uart_port.state->port.tty, ch, flag);
#endif
}

static inline void aslser_tty_flip_buffer_push(struct aslser_device *ser_device)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,9,0)
   tty_flip_buffer_push(&ser_device->uart_port.state->port);
#else
   tty_flip_buffer_push(ser_device->uart_port.state->port.tty);
#endif
   return;
}

/* Called with aslser spinlock held */
static void fbclser_rx_isr(struct aslser_device *ser_device)
{
   uint32_t ulStatus = 0;
   uint32_t ulBytesInFifo = 0;
   uint32_t ulBytesReceived = 0;

   ulStatus = mda_read32(ser_device->mdadev, FBD_SERIAL_OFFSET_STAT);
   ulBytesInFifo = (ulStatus & FBD_SERIAL_BITS_STAT_RX_FIFO_MASK) >> FBD_SERIAL_BITS_STAT_RX_FIFO_SHIFT;

   while(ulBytesInFifo-- > 0)
   {
      uint32_t ulFifoData = mda_read32(ser_device->mdadev, FBD_SERIAL_OFFSET_RX_DATA);
      char tty_data = (char)(ulFifoData & ser_device->valid_data_mask);


      if (!ser_device->enable_rx){
         continue; /* Drop the character */
      }
      ser_device->uart_port.icount.rx++;
      if (ulFifoData & FBD_SERIAL_BITS_RX_ERROR_BREAK){
         aslser_tty_insert_flip_char(ser_device, 0, TTY_BREAK);
         ulBytesReceived++;
         continue;
      }
      if (ulFifoData & FBD_SERIAL_BITS_RX_ERROR_FRAME){
         aslser_tty_insert_flip_char(ser_device, 0, TTY_FRAME);
         ulBytesReceived++;
         if (ser_device->enable_frame_error) ser_device->uart_port.icount.frame++;
         continue;
      }
      if (ulFifoData & FBD_SERIAL_BITS_RX_ERROR_PARITY){
         if (ser_device->enable_parity_error) ser_device->uart_port.icount.parity++;
         if (ser_device->enable_parity_checking){
            continue; /* drop the character */
         }
         aslser_tty_insert_flip_char(ser_device, tty_data, ser_device->enable_parity_error ? TTY_PARITY : TTY_NORMAL);
         ulBytesReceived++;
         continue;
      }
      aslser_tty_insert_flip_char(ser_device, tty_data, TTY_NORMAL);
      ulBytesReceived++;
   }

   if (ulStatus & FBD_SERIAL_BITS_STAT_RX_FIFO_OVERRUN){
      mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control | FBD_SERIAL_BITS_CTRL_RX_OVERRUN_CLEAR);
      mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control);
      ser_device->uart_port.icount.overrun++;
      aslser_tty_insert_flip_char(ser_device, 0, TTY_OVERRUN);
      ulBytesReceived++;
   }

   if (ulBytesReceived) aslser_tty_flip_buffer_push(ser_device);

   return;
}

/* Called with aslser spinlock held */
static void fbclser_tx_isr(struct aslser_device *ser_device)
{
   uint32_t ulStatus = 0;
   uint32_t ulSpaceInFifo = 0;

   ulStatus = mda_read32(ser_device->mdadev, FBD_SERIAL_OFFSET_STAT);
   ulSpaceInFifo = ulStatus & FBD_SERIAL_BITS_STAT_TX_FIFO_MASK;

   if (ASLSER_TX_IMMEDIATE_VALID(ser_device->tx_char_flag_and_data)){ /* Takes precedence, even in the middle of another write operation */
      uint32_t ulCharacter = ASLSER_TX_IMMEDIATE_CHAR(ser_device->tx_char_flag_and_data);
      ulCharacter &= ser_device->valid_data_mask;
      mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_TX_DATA, ulCharacter);
      ser_device->tx_char_flag_and_data = 0;
      ser_device->uart_port.icount.tx++;
      goto bail;
   }

   if (uart_tx_stopped(&ser_device->uart_port)){
      fbclser_tx_off(ser_device);
      goto bail;
   }

   if (uart_circ_empty(&ser_device->uart_port.state->xmit)){
      fbclser_tx_off(ser_device);
      goto bail;
   }

   while((!uart_circ_empty(&ser_device->uart_port.state->xmit)) && (ulSpaceInFifo > 0)){
      int tail = ser_device->uart_port.state->xmit.tail;
      char tx_char = ser_device->uart_port.state->xmit.buf[tail];
      uint32_t tx_value = (uint32_t)tx_char & ser_device->valid_data_mask;
      tail = (tail + 1) & (UART_XMIT_SIZE -1);
      mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_TX_DATA, tx_value);
      ser_device->uart_port.state->xmit.tail = tail;
      ulSpaceInFifo--;
      ser_device->uart_port.icount.tx++;
   }

   if (uart_circ_chars_pending(&ser_device->uart_port.state->xmit) < WAKEUP_CHARS){
      uart_write_wakeup(&ser_device->uart_port);
   }

   ulStatus = mda_read32(ser_device->mdadev, FBD_SERIAL_OFFSET_STAT);
   ulSpaceInFifo = ulStatus & FBD_SERIAL_BITS_STAT_TX_FIFO_MASK;
bail:
    return;
}

static irqreturn_t fbclser_isr(struct mda_device *mda_device, void *data)
{
   struct aslser_device * ser_device = (struct aslser_device*)data;
   uint32_t interrupt_status = 0;

   spin_lock(&ser_device->uart_port.lock);
   interrupt_status = mda_read32(ser_device->mdadev, FBD_SERIAL_OFFSET_INT_STAT);


   if (interrupt_status & FBD_SERIAL_BITS_INT_STAT_RXDATA_ASSERTED){
      fbclser_rx_isr(ser_device);
   }
   if (interrupt_status & (FBD_SERIAL_BITS_INT_STAT_TXEMPTY_ASSERTED | FBD_SERIAL_BITS_INT_STAT_TXHALF_ASSERTED)){
      fbclser_tx_isr(ser_device);
   }
   spin_unlock(&ser_device->uart_port.lock);
   return IRQ_HANDLED;
}

/* called unlocked */
static unsigned int aslser_fbcl_tx_empty(struct uart_port *uart_port)
{
   unsigned int empty = 0;
   struct aslser_device *ser_device = to_mda_ser(uart_port);
   unsigned long irq_state = 0;
   uint32_t interrupt_status = 0;

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);

   spin_lock_irqsave(&ser_device->uart_port.lock, irq_state);
   interrupt_status = mda_read32(ser_device->mdadev, FBD_SERIAL_OFFSET_INT_STAT);
   if (interrupt_status & FBD_SERIAL_BITS_INT_STAT_TXEMPTY_PENDING) empty = TIOCSER_TEMT;
   spin_unlock_irqrestore(&ser_device->uart_port.lock, irq_state);

   return empty;
}

/* called locked */
static void aslser_fbcl_set_mctrl(struct uart_port *uart_port, unsigned int mctrl)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);
   /* None connected on current FireBirds */
   return;
}

/* called locked */
static unsigned int aslser_fbcl_get_mctrl(struct uart_port *uart_port)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);
   /* Fake - none connected on current FireBirds */
   return TIOCM_CTS | TIOCM_DSR | TIOCM_CAR;
}

/* called locked */
static void aslser_fbcl_stop_tx(struct uart_port *uart_port)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);

   fbclser_tx_off(ser_device);

   return;
}

/* called locked */
static void aslser_fbcl_start_tx(struct uart_port *uart_port)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);

   fbclser_tx_on(ser_device);

   return;
}

/* called locked */
static void aslser_fbcl_send_xchar(struct uart_port *uart_port, char ch)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_printk(ser_device, KERN_INFO, "%s: 0x%02X\n", __FUNCTION__, ch);

   ser_device->tx_char_flag_and_data = ((uint32_t)ch) | ASLSER_TX_IMMEDIATE_FLAG;
   fbclser_tx_on(ser_device);

   return;
}

/* called locked */
static void aslser_fbcl_stop_rx(struct uart_port *uart_port)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);

   ser_device->shadow_interrupt_control &= ~FBD_SERIAL_BITS_INT_CTRL_RXDATA_ENABLE;
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_INT_CTRL, ser_device->shadow_interrupt_control);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control | FBD_SERIAL_BITS_CTRL_RX_FIFO_RESET);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control);

   return;
}

/* called locked */
static void aslser_fbcl_enable_ms(struct uart_port *uart_port)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);
   /* None connected */
   return;
}

/* Called unlocked */
static void aslser_fbcl_break_ctl(struct uart_port *uart_port, int ctl)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);
   unsigned long irq_state = 0;

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);

   spin_lock_irqsave(&ser_device->uart_port.lock, irq_state);
   if (ctl) ser_device->shadow_serial_control |= FBD_SERIAL_BITS_CTRL_BREAK;
   else ser_device->shadow_serial_control &= ~FBD_SERIAL_BITS_CTRL_BREAK;
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control);
   spin_unlock_irqrestore(&ser_device->uart_port.lock, irq_state);
   return;
}

/* called unlocked, but port_sem taken*/
static int aslser_fbcl_startup(struct uart_port *uart_port)
{
   int error = -EIO;
   struct aslser_device *ser_device = to_mda_ser(uart_port);
   unsigned long irq_state = 0;

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);

   aslser_fbcl_reset_port(ser_device);

   error = mda_interrupt_register(ser_device->mdadev,
                                  ser_device->mdadev->params.interrupt_register,
                                  ser_device->mdadev->params.interrupt_mask,
                                  fbclser_isr,
                                  ser_device);
   if (error){
      printk(KERN_ERR "%s: mda_interrupt_register returned %d\n", __FUNCTION__, error);
      goto bail;
   }

   spin_lock_irqsave(&ser_device->uart_port.lock, irq_state);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control | FBD_SERIAL_BITS_CTRL_RX_FIFO_RESET);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control | FBD_SERIAL_BITS_CTRL_TX_FIFO_RESET);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control);
   ser_device->shadow_interrupt_control |= FBD_SERIAL_BITS_INT_CTRL_RXDATA_ENABLE;
   ser_device->shadow_interrupt_control |= FBD_SERIAL_BITS_INT_CTRL_GLOBAL_ENABLE;
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_INT_CTRL, ser_device->shadow_interrupt_control);
   spin_unlock_irqrestore(&ser_device->uart_port.lock, irq_state);

bail:
   return error;
}

/* called unlocked, but port_sem taken*/
static void aslser_fbcl_shutdown(struct uart_port *uart_port)
{
   int error = 0;
   struct aslser_device *ser_device = to_mda_ser(uart_port);
   unsigned long irq_state = 0;

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);

   spin_lock_irqsave(&ser_device->uart_port.lock, irq_state);
   ser_device->shadow_interrupt_control = 0;
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_INT_CTRL, ser_device->shadow_interrupt_control);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control | FBD_SERIAL_BITS_CTRL_RX_FIFO_RESET);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control | FBD_SERIAL_BITS_CTRL_TX_FIFO_RESET);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control);
   spin_unlock_irqrestore(&ser_device->uart_port.lock, irq_state);

   aslser_fbcl_reset_port(ser_device);

   error = mda_interrupt_unregister(ser_device->mdadev,
                                    ser_device->mdadev->params.interrupt_register,
                                    ser_device->mdadev->params.interrupt_mask);
   if (error){
      aslser_fbcl_printk(ser_device, KERN_ERR, "%s: mda_interrupt_unregister returned %d\n", __FUNCTION__, error);
   }
   return;
}

/* called locked */
static void aslser_fbcl_flush_buffer(struct uart_port *uart_port)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);
   /* No action required */
   return;
}

/* called unlocked */
static void aslser_fbcl_set_termios(struct uart_port *uart_port, struct ktermios *newconf, struct ktermios *oldconf)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);
   unsigned long irq_state = 0;
   unsigned int baudrate = 0;

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);

   spin_lock_irqsave(&ser_device->uart_port.lock, irq_state);

   ser_device->enable_break_error = 0;
   ser_device->enable_frame_error = 0;
   ser_device->enable_parity_error = 0;
   ser_device->enable_rx = 0;
   ser_device->shadow_serial_control = 0;
   ser_device->enable_parity_checking = 0;

   if (newconf->c_iflag & INPCK){
      ser_device->enable_frame_error = 1;
      ser_device->enable_parity_error = 1;
      ser_device->enable_parity_checking = 1;
   }
   if (newconf->c_iflag & IGNPAR) ser_device->enable_parity_error = 0;

   if (newconf->c_cflag & CREAD) ser_device->enable_rx = 1;

   if ((newconf->c_cflag & CSIZE) == CS5){
      ser_device->valid_data_mask = 0x1F;
      ser_device->shadow_serial_control |= FBD_SERIAL_BITS_CTRL_DATA_BITS_5;
   }
   else if ((newconf->c_cflag & CSIZE) == CS6){
      ser_device->valid_data_mask = 0x3F;
      ser_device->shadow_serial_control |= FBD_SERIAL_BITS_CTRL_DATA_BITS_6;
   }
   else if ((newconf->c_cflag & CSIZE) == CS7){
      ser_device->valid_data_mask = 0x7F;
      ser_device->shadow_serial_control |= FBD_SERIAL_BITS_CTRL_DATA_BITS_7;
   }
   else{
      ser_device->valid_data_mask = 0xFF;
      ser_device->shadow_serial_control |= FBD_SERIAL_BITS_CTRL_DATA_BITS_8;
   }

   if ((newconf->c_cflag & CSTOPB) == CSTOPB){
      ser_device->shadow_serial_control |= FBD_SERIAL_BITS_CTRL_STOP_2;
   }
   else{
      ser_device->shadow_serial_control |= FBD_SERIAL_BITS_CTRL_STOP_1;
   }

   if ((newconf->c_cflag & PARENB) == PARENB){
      if ((newconf->c_cflag & PARODD) == PARODD){
         ser_device->shadow_serial_control |= FBD_SERIAL_BITS_CTRL_PARITY_ODD;
      }
      else{
         ser_device->shadow_serial_control |= FBD_SERIAL_BITS_CTRL_PARITY_EVEN;
      }
   }
   else{
      ser_device->shadow_serial_control |= FBD_SERIAL_BITS_CTRL_PARITY_NONE;
   }

   baudrate = uart_get_baud_rate(uart_port, newconf, oldconf, 0, uart_port->uartclk/16);
   ser_device->shadow_baud_control = uart_get_divisor(uart_port, baudrate);
   uart_update_timeout(uart_port, newconf->c_cflag, baudrate);

   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_BAUD, ser_device->shadow_baud_control);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control | FBD_SERIAL_BITS_CTRL_TX_FIFO_RESET);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control | FBD_SERIAL_BITS_CTRL_RX_FIFO_RESET);
   mda_write32(ser_device->mdadev, FBD_SERIAL_OFFSET_CTRL, ser_device->shadow_serial_control);

   spin_unlock_irqrestore(&ser_device->uart_port.lock, irq_state);
   aslser_fbcl_printk(ser_device,
                      KERN_INFO,
                      "%s: Baud = %d, Divisor = %d, control = 0x%08X\n",
                      __FUNCTION__,
                      baudrate,
                      ser_device->shadow_baud_control,
                      ser_device->shadow_serial_control);
   return;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0)
static void aslser_fbcl_set_ldisc(struct uart_port *uart_port, struct ktermios *kt)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,38)
static void aslser_fbcl_set_ldisc(struct uart_port *uart_port, int new)
#else
static void aslser_fbcl_set_ldisc(struct uart_port *uart_port)
#endif
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_printk(ser_device, KERN_INFO, "%s", __FUNCTION__);

   return;
}

/* called unlocked */
static void aslser_fbcl_pm(struct uart_port *uart_port, unsigned int state, unsigned int oldstate)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_printk(ser_device, KERN_INFO, "%s: Moving from pm %d to %d\n", __FUNCTION__, oldstate, state);

   return;
}

/* Not supported */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,13,0)
static int  aslser_fbcl_set_wake(struct uart_port *uart_port, unsigned int state)
{
   unsigned int error = -EINVAL;
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_printk(ser_device, KERN_INFO, "%s: state %d\n", __FUNCTION__, state);

   return error;
}
#endif

/* called unlocked */
static const char * aslser_fbcl_type(struct uart_port *uart_port)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);

   return "Firebird CameraLink Port";
}


/* called unlocked */
static void aslser_fbcl_release_port(struct uart_port *uart_port)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);
   aslser_fbcl_reset_port(ser_device);

   return;
}

/* called unlocked */
static int  aslser_fbcl_request_port(struct uart_port *uart_port)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);
   aslser_fbcl_reset_port(ser_device);

   return 0;
}

/* called unlocked */
static void aslser_fbcl_config_port(struct uart_port *uart_port, int flags)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);

   if (flags & UART_CONFIG_TYPE) {
      if (aslser_fbcl_request_port(uart_port)){
         goto bail;
      }
      uart_port->type = PORT_UARTLITE;
      aslser_fbcl_printk(ser_device, KERN_INFO, "%s: Setting type to PORT_UARTLITE\n", __FUNCTION__);
   }


bail:
   return;
}

/* called unlocked */
static int aslser_fbcl_verify_port(struct uart_port *uart_port, struct serial_struct * serial_in)
{
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_debug_entry(ser_device, __FUNCTION__);

   return -EINVAL; /* Modifications not allowed */
}

#if 0
/* called unlocked */
static int  aslser_fbcl_ioctl(struct uart_port *uart_port, unsigned int arg1, unsigned long arg2)
{
   unsigned int error = -EINVAL;
   struct aslser_device *ser_device = to_mda_ser(uart_port);

   aslser_fbcl_printk(ser_device, KERN_INFO, "%s: %08X (%p)\n", __FUNCTION__, arg1, (void*)arg2);

   return error;
}
#endif

struct uart_ops fbclser_ops = {
   .tx_empty = aslser_fbcl_tx_empty,
   .set_mctrl = aslser_fbcl_set_mctrl,
   .get_mctrl = aslser_fbcl_get_mctrl,
   .stop_tx = aslser_fbcl_stop_tx,
   .start_tx = aslser_fbcl_start_tx,
   .send_xchar = aslser_fbcl_send_xchar,
   .stop_rx = aslser_fbcl_stop_rx,
   .enable_ms = aslser_fbcl_enable_ms,
   .break_ctl = aslser_fbcl_break_ctl,
   .startup = aslser_fbcl_startup,
   .shutdown = aslser_fbcl_shutdown,
   .flush_buffer = aslser_fbcl_flush_buffer,
   .set_termios = aslser_fbcl_set_termios,
   .set_ldisc = aslser_fbcl_set_ldisc,
   .pm = aslser_fbcl_pm,
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,13,0)
   .set_wake = aslser_fbcl_set_wake,
#endif
   .type = aslser_fbcl_type,
   .release_port = aslser_fbcl_release_port,
   .request_port = aslser_fbcl_request_port,
   .config_port = aslser_fbcl_config_port,
   .verify_port = aslser_fbcl_verify_port,
#if 0
   .ioctl = aslser_fbcl_ioctl
#endif
};

