/****************************************************************************
 *
 * ACTIVE SILICON LIMITED
 *
 * File name   : mda_all.h
 * Function    : Common MDA Kernel Interface.
 * Project     : MDA
 * Authors     : Martin Bone
 * Systems     : ANSI C
 *
 * Copyright (c) 1998-2012 Active Silicon.
 ****************************************************************************
 * Comments:
 * --------
 * Note this include file is also used by drivers.
 *
 ****************************************************************************
 */

#ifndef _MDA_ALL_H_
#define _MDA_ALL_H_


/*
 ****************************************************************************
 * MDA_BUF
 ****************************************************************************
 * Scatter/Gather Table (Physical Addresses & Lengths)
 * ---------------------------------------------------
 * Scatter/Gather table information/macros for locking and freeing physical
 * memory. Note the Area For Transfer (AFT) entries - this allows a sub-area
 * within the total scatter/gather table to be handled. This can be useful to
 * handle a region of interest within a larger region.
 * 
 * The SG table is of the form (all entries are ui32s):
 *
 *   pSGTable-> 0. Allocated size of table (in bytes),
 *              1. Virtual pointer (to base address),
 *              2. Number of Entries (Physical address/length pairs),
 *              3. Total amount of memory represented by table (in bytes),
 *              4. Offset (in bytes - virtual) to Area For Transfer (AFT),
 *              5. Transfer Size of AFT (if 0, then transfer full area).
 *              - Address,
 *              - Length,
 *              ...
 *
 * x = Pointer to SG table, y = Entry number (0..N-1), z = Value to set.
 * 
 * All sizes are in bytes.
 *
 */

/* Use 64 bit Scatter-Gather tables by default */
#if defined _MDA_SG32
  typedef uint32_t uiSG_t;
  typedef  int32_t uiSGOffset_t;
#else
  typedef uint64_t uiSG_t;
  typedef  int64_t uiSGOffset_t;
#endif


typedef struct {
   uiSG_t   uiPhysAddr;
   uiSG_t   uiPhysLen;
} MDA_SgEntry, *PMDA_SgEntry;


typedef struct {
   uiSG_t   uiAllocatedSize;
   uiSG_t   uiVirtualAddr;
   uiSG_t   uiNumEntries;
   uiSG_t   uiMemorySize;
   uiSG_t   uiStartOffset;
   uiSG_t   uiTransferSize;
} MDA_SgHeader, *PMDA_SgHeader;

typedef struct {
   MDA_SgHeader    sSgHeader;
   PMDA_SgEntry   pSgEntries;
} MDA_SgInfo, *PMDA_SgInfo;


#define MDA_SGTABLE_HEADER_SIZE      ((uiSG_t) sizeof(MDA_SgHeader))
#define MDA_SGTABLE_BYTES_PER_ENTRY  ((uiSG_t) sizeof(MDA_SgEntry))

#define _MDA_SGTable_Get_AllocatedSize(x)     (((PMDA_SgInfo)(x))->sSgHeader.uiAllocatedSize)
#define _MDA_SGTable_Get_VirtualAddr(x)       (((PMDA_SgInfo)(x))->sSgHeader.uiVirtualAddr)
#define _MDA_SGTable_Get_NumEntries(x)        (((PMDA_SgInfo)(x))->sSgHeader.uiNumEntries)
#define _MDA_SGTable_Get_MemorySize(x)        (((PMDA_SgInfo)(x))->sSgHeader.uiMemorySize)
#define _MDA_SGTable_Get_StartOffset(x)       (((PMDA_SgInfo)(x))->sSgHeader.uiStartOffset)
#define _MDA_SGTable_Get_TransferSize(x)      (((PMDA_SgInfo)(x))->sSgHeader.uiTransferSize)

#define _MDA_SGTable_Get_PhysAddr(x,y)        (((PMDA_SgInfo)(x))->pSgEntries[y].uiPhysAddr)
#define _MDA_SGTable_Get_PhysLen(x,y)         (((PMDA_SgInfo)(x))->pSgEntries[y].uiPhysLen)

#define _MDA_SGTable_Set_AllocatedSize(x,z)   (((PMDA_SgInfo)(x))->sSgHeader.uiAllocatedSize = z)
#define _MDA_SGTable_Set_VirtualAddr(x,z)     (((PMDA_SgInfo)(x))->sSgHeader.uiVirtualAddr   = z)
#define _MDA_SGTable_Set_NumEntries(x,z)      (((PMDA_SgInfo)(x))->sSgHeader.uiNumEntries    = z)
#define _MDA_SGTable_Set_MemorySize(x,z)      (((PMDA_SgInfo)(x))->sSgHeader.uiMemorySize    = z)
#define _MDA_SGTable_Set_StartOffset(x,z)     (((PMDA_SgInfo)(x))->sSgHeader.uiStartOffset   = z)
#define _MDA_SGTable_Set_TransferSize(x,z)    (((PMDA_SgInfo)(x))->sSgHeader.uiTransferSize  = z)

#define _MDA_SGTable_AddrLo32(x) ((ui32) (x)    )
#if defined _MDA_SG32
  /* For library compatability with 32 bit only OSs, redefine the Hi & Lo addr macros */
  #define _MDA_SGTable_AddrHi32(x) ((ui32) (0))
#else
#define _MDA_SGTable_AddrHi32(x) ((ui32) ((x)>>32))
#endif



/****************************************************************************
 * MDA Register Lists
 ****************************************************************************
 */

#define CDA_LIB_ID            ((uint32_t) 0xCD000000)
#define CDA_MAGIC_REG_TYPE    ((uint32_t) CDA_LIB_ID | ( 1 << 16))

#define CDA_8BIT              (CDA_MAGIC_REG_TYPE | (uint32_t) 0x0001)
#define CDA_16BIT             (CDA_MAGIC_REG_TYPE | (uint32_t) 0x0002)
#define CDA_32BIT             (CDA_MAGIC_REG_TYPE | (uint32_t) 0x0004)

/* For clarity and convenience, CDA_NO_OP is defined here, although it is 
 * not used by the user. It is used internally to indicate the end of a
 * a cached register list. 
 */
#define CDA_NO_OP         (CDA_MAGIC_REG_TYPE | (uint32_t) 0x0008) 
#define CDA_NOW           (CDA_MAGIC_REG_TYPE | (uint32_t) 0x0010)
#define CDA_CACHED        (CDA_MAGIC_REG_TYPE | (uint32_t) 0x0020) /* If clear assume CDA_NOW */
#define CDA_READ_NO_FLUSH (CDA_MAGIC_REG_TYPE | (uint32_t) 0x0040) /* Allows reading without flushing */
#define CDA_MODE_MASK     (                     (uint32_t) 0x0070)


/* CDA Register Types
 */
#define CDA_REG_WRITE      ((uint32_t) 0x00000100)
#define CDA_REG_READ       ((uint32_t) 0x00000200)
#define CDA_CNF_READ       ((uint32_t) 0x00000400)
#define CDA_CNF_WRITE      ((uint32_t) 0x00000800)

#define CDA_REG_WRITE_8BITS    ( CDA_REG_WRITE | CDA_8BIT  )
#define CDA_REG_WRITE_16BITS   ( CDA_REG_WRITE | CDA_16BIT )
#define CDA_REG_WRITE_32BITS   ( CDA_REG_WRITE | CDA_32BIT )

#define CDA_REG_READ_8BITS     ( CDA_REG_READ | CDA_8BIT  )
#define CDA_REG_READ_16BITS    ( CDA_REG_READ | CDA_16BIT )
#define CDA_REG_READ_32BITS    ( CDA_REG_READ | CDA_32BIT )

#define CDA_CNF_WRITE_8BITS    ( CDA_CNF_WRITE | CDA_8BIT  )
#define CDA_CNF_WRITE_16BITS   ( CDA_CNF_WRITE | CDA_16BIT )
#define CDA_CNF_WRITE_32BITS   ( CDA_CNF_WRITE | CDA_32BIT )

#define CDA_CNF_READ_8BITS     ( CDA_CNF_READ | CDA_8BIT  )
#define CDA_CNF_READ_16BITS    ( CDA_CNF_READ | CDA_16BIT )
#define CDA_CNF_READ_32BITS    ( CDA_CNF_READ | CDA_32BIT )


/* Define a container for the register accesses requests
 */
typedef struct {
   uint32_t    eMdaRegType;   /* Access type, ie Read or Write */
   uint32_t    uiAddress;     /* Address to Read or Write */
   uint32_t    uiData;        /* Data to Write */
} CDA_RegReq, *PCDA_RegReq;
#define CDA_REG_RECORD_SIZE sizeof( CDA_RegReq )
#define _CDA_RegList_GetRegType(p)      (((PCDA_RegReq)(p))->eMdaRegType)
#define _CDA_RegList_GetRegAddr(p)      (((PCDA_RegReq)(p))->uiAddress)
#define _CDA_RegList_GetRegData32(p)    (((PCDA_RegReq)(p))->uiData)
#define _CDA_RegList_SetRegData32(p, d) (((PCDA_RegReq)(p))->uiData = d)

/****************************************************************************
 * DeviceInfo Structure
 ****************************************************************************
 */
#define CDA_kDevicePci  1
#define CDA_kDeviceIsa  2
typedef struct {
   uint32_t dwDriverRev;         /* b0..15 minor, b16..31 major */
   uint32_t type;                /* Bus type: CDA_kDevice... */
   uint16_t wDeviceID;
   uint16_t wVendorID;
   uint16_t wPciSubVendorID;
   uint16_t wPciSubSystemID;
   uint16_t wDeviceNumber;       /* PCI slot no. */
   uint16_t wFunctionNumber;     /* PCI functional unit */
   uint32_t dwBase;              /* Bus logical base address */
   uint32_t dwInterrupt;
} CDA_IoctlDeviceInfo, *PCDA_IoctlDeviceInfo;



/****************************************************************************
 * Event Structures
 ****************************************************************************
 */
/* Original CDA Event structure
 */
typedef struct {
   uint32_t dwEvent;             /* Device specific event ID e.g. LFG irq status */
   uint32_t dwData;              /* Event specific data */
} CDA_IoctlEvent, *PCDA_IoctlEvent;

#define CDA_IOCTL_EVENT_DRV_EVENT         0x80000000
#define CDA_IOCTL_EVENT_DATA_CUDA_FREE    0x00000001


/* Enhanced MDA Event structure. Conveys the same information
 * as the standard CDA Event structure, plus timing and debugging
 * information. Useful for debugging systems with high rate
 * event signalling.
 */
typedef struct {
   uint64_t       qwSize;
   uint64_t       qwFlags;
   uint64_t       qwSequenceNumber;
   uint64_t       qwInterruptTimestamp;
   uint64_t       qwDispatchTimestamp;
   CDA_IoctlEvent sEvent;
}  MDA_IoctlEvent, *PMDA_IoctlEvent;


/****************************************************************************
 * ASL_CXP Structures
 ****************************************************************************
 */
typedef struct {
   uint64_t qwHardwareTimeStamp;
   uint64_t qwInterruptTimeStamp;
   uint64_t qwDispatchTimeStamp;
   uint64_t qwUserTimeStamp;
} MDA_CxpTiming, *PMDA_CxpTiming;

typedef struct {
   uint32_t dwDeviceAddress;
   uint32_t dwByteCount;
} MDA_CxpDataHeader, *PMDA_CxpDataHeader;

typedef struct {
   MDA_CxpDataHeader sHeader;
   uint8_t           baData[1];
} MDA_CxpData, *PMDA_CxpData;

typedef struct {
  uint32_t  dwGpioValue;
  uint32_t  dwGpioStatus;
  MDA_CxpTiming  sTimingInfo;
} MDA_CxpGpioEvent, *PMDA_CxpGpioEvent;

typedef struct {
   uint32_t dwOffset;
   uint32_t dwData;
} MDA_CxpRegister, *PMDA_CxpRegister;


/****************************************************************************
 * ASL_DMA Structures
 ****************************************************************************
 */
typedef struct {
   uint64_t qwVirtualAddress;
   uint64_t qwSize;
} MDA_IoctlMemory, *PMDA_IoctlMemory;


typedef struct {
   uint64_t qwMaxMdl;
   uint64_t qwMaxDmaTotal;
   uint64_t qwMaxReadFragment;
   uint64_t qwMaxWriteFragment;
   uint64_t qwMaxTransfer;
   uint64_t qwMaxSGElements;
} MDA_DmaInfo, *PMDA_DmaInfo;


/****************************************************************************
 * ASL_ENUM Structures
 ****************************************************************************
 */
#define MDAENUM_DEVICE_NAME_REQUEST_TYPE_UNKNOWN  0
#define MDAENUM_DEVICE_NAME_REQUEST_TYPE_DMA      1
#define MDAENUM_DEVICE_NAME_REQUEST_TYPE_CXP      2
#define MDAENUM_DEVICE_NAME_REQUEST_TYPE_SERIAL   3
#define MDAENUM_DEVICE_NAME_REQUEST_TYPE_GIO      4
#define MDAENUM_DEVICE_NAME_REQUEST_TYPE_SERRAW   5
#define MDAENUM_DEVICE_NAME_REQUEST_TYPE_CANBUS   6

/* Get child device name */
typedef struct {
   uint32_t    uiChannelIndex;
   uint32_t    uiDeviceIndex;
   uint32_t    uiType;
} MDA_DeviceNameRequest, *PMDA_DeviceNameRequest;


/****************************************************************************
 * ASL_CAN Structures
 ****************************************************************************
 */
#define MDACAN_MAX_PAYLOAD_SIZE     8
#define MDACAN_FLAGS_RX_OVERFLOW    1
typedef struct {
   uint32_t uiAddress;
   uint32_t fRTR;
   uint32_t uiNumBytes;
   uint32_t uiFlags;
   uint8_t  baPayload[MDACAN_MAX_PAYLOAD_SIZE];
} MDA_IoctlCanBusPacket, *PMDA_IoctlCanBusPacket;

#define _CAN_MSSG_PAYLOAD_MAX     MDACAN_MAX_PAYLOAD_SIZE
#define _CAN_MSSG_HEADER_SIZE    (sizeof( MDA_IoctlCanBusPacket ) - MDACAN_MAX_PAYLOAD_SIZE)


typedef struct {
   uint32_t       uiCanMask;
   uint32_t       uiCanFilter;
   uint32_t       uiBaudRate;
   uint32_t       fExtended;
} MDA_IoctlCanParameters, *PMDA_IoctlCanParameters;



/****************************************************************************
 * Power Events
 ****************************************************************************
 */
#define MDA_POWER_REASON_NONE          0
#define MDA_POWER_REASON_DMA_STOP      1
#define MDA_POWER_REASON_DMA_START     2
#define MDA_POWER_REASON_POWER_OFF     3
#define MDA_POWER_REASON_POWER_ON      4

typedef struct {
    uint32_t   eReason;
    uint32_t   dwInformation;
    uint64_t   qwSequenceNumber;
    uint64_t   qwSignalTimestamp;
    uint64_t   qwDispatchTimestamp;
} MDA_IoctlPowerEvent, *PMDA_IoctlPowerEvent;


/****************************************************************************
 * IOCTL Buffers
 ****************************************************************************
 */
typedef struct {
   void * user_address;
   size_t user_size;
} mda_user_buffer, *pmda_user_buffer;

typedef struct {
   mda_user_buffer param_buffer;
   mda_user_buffer input_buffer;
   mda_user_buffer output_buffer;
} mda_io_buffer, *pmda_io_buffer;


/****************************************************************************
 * Constant Definitions
 ****************************************************************************
 */
#define MDALOCK_TIMEOUT_IMMEDIATE                     0
#define MDALOCK_TIMEOUT_INFINITE                      0xFFFFFFFF
#define MDA_IOCTL_CAN_FLAG_INTERRUPT_QUEUE_OVERRUN    0x8000000000000002ULL
#define MDA_IOCTL_CAN_FLAG_INTERRUPT_FIFO_OVERRUN     0x8000000000000001ULL
#define MDA_IOCTL_MDA_FLAG_INTERRUPT_QUEUE_OVERRUN    0x8000000000000002ULL
#define MDA_IOCTL_MDA_FLAG_INTERRUPT_FIFO_OVERRUN     0x8000000000000001ULL
#define MDA_DMAINFO_UNLIMITED                         0xFFFFFFFFFFFFFFFCULL
#define MDA_DMAINFO_UNLIMITED_32                      0x00000000FFFFFFFC

/* CXP Error Returns
 */
#define MDA_CXP_DEVICE_ERROR                          0x00000000
#define MDA_CXP_DRIVER_ERROR                          0x80000000
#define MDA_CXP_HOST_ERROR                            0x90000000

#define MDA_CXP_ACK_READ_OK                           0x00
#define MDA_CXP_ACK_WRITE_OK                          0x01
#define MDA_CXP_TENTATIVE_ACK                         0x02
#define MDA_CXP_RST_OK                                0x03
#define MDA_CXP_WAIT_ACK                              0x04
#define MDA_CXP_ERROR_INVALID_ADDRESS                 0x40
#define MDA_CXP_ERROR_INVALID_DATA                    0x41
#define MDA_CXP_ERROR_INVALID_CONTROL                 0x42
#define MDA_CXP_ERROR_INVALID_WRITE                   0x43
#define MDA_CXP_ERROR_INVALID_READ                    0x44
#define MDA_CXP_ERROR_SIZE_TOO_LARGE                  0x45
#define MDA_CXP_ERROR_INVALID_SIZE                    0x46
#define MDA_CXP_ERROR_MALFORMED_PACKET                0x47
#define MDA_CXP_ERROR_INCORRECT_CRC                   0x80

#define MDA_CXP_ERROR_SIZE_MISMATCH                   ( MDA_CXP_DRIVER_ERROR | MDA_CXP_ERROR_INVALID_SIZE    )
#define MDA_CXP_ERROR_STATUS_MISMATCH                 ( MDA_CXP_DRIVER_ERROR | MDA_CXP_ERROR_INCORRECT_CRC   )
#define MDA_CXP_ERROR_COMMAND_MISMATCH                ( MDA_CXP_DRIVER_ERROR | MDA_CXP_ERROR_INVALID_CONTROL )
#define MDA_CXP_ERROR_WAIT_ACK_DATA                   ( MDA_CXP_DRIVER_ERROR | MDA_CXP_ERROR_SIZE_TOO_LARGE  )

#define MDA_CXP_ERROR_ACK_TIMEOUT                     ( MDA_CXP_DRIVER_ERROR | 0x00000100 )
#define MDA_CXP_ERROR_WAIT_ACK_TIMEOUT                ( MDA_CXP_DRIVER_ERROR | 0x00000200 )
#define MDA_CXP_ERROR_USER_ACK_TIMEOUT                ( MDA_CXP_DRIVER_ERROR | 0x00000400 )
#define MDA_CXP_ERROR_INITIAL_ACK_TIMEOUT             ( MDA_CXP_DRIVER_ERROR | 0x00000800 )
#define MDA_CXP_ERROR_TENTATIVE_ACK_TIMEOUT           ( MDA_CXP_DRIVER_ERROR | 0x00001000 )

#define MDA_CXP_ERROR_RX_PACKET_INVALID               ( MDA_CXP_HOST_ERROR   | MDA_CXP_ERROR_INVALID_DATA  )
#define MDA_CXP_ERROR_RX_PACKET_CRC_ERROR             ( MDA_CXP_HOST_ERROR   | MDA_CXP_ERROR_INCORRECT_CRC )


#endif /* _MDA_ALL_H_ */
